# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Additional configuration settings for experimental firmware features

Any scripts in firmware/experimental that need editiable, persistent settings should use this. If/when
they're moved out of experimental, their associated settings should be migrated to europi_config.py
"""

import configuration
from configuration import ConfigFile, ConfigSpec

# Moog/Eurorack standard is 1.0 volts per octave
MOOG_VOLTS_PER_OCTAVE = 1.0

# Buchla standard is 1.2 volts per octave (0.1 volts per semitone)
BUCHLA_VOLTS_PER_OCTAVE = 1.2

# RTC implementations
RTC_NONE = ""
RTC_DS3231 = "ds3231"
RTC_DS1307 = "ds1307"
RTC_NTP = "ntp"

# WIFI modes
WIFI_MODE_AP = "access_point"
WIFI_MODE_CLIENT = "client"


class ExperimentalConfig:
    """This class provides global config points for experimental features.

    To override the default values, create /config/config_ExperimentalConfig.json on the Raspberry Pi Pico
    and populate it with a JSON object. See CONFIGURATION.md for details on available keys.
    """

    @classmethod
    def config_points(cls):
        """
        Get the array of configuration points for EuroPi's experimental libraries

        :param cls: Our own class
        """
        # fmt: off
        return [
            # Quantizer settings
            # Normally this is intended for Eurorack compatibility, but being open-source someone may
            # want to use it in an ecosystem that uses different specs
            configuration.choice(
                "VOLTS_PER_OCTAVE",
                choices=[MOOG_VOLTS_PER_OCTAVE, BUCHLA_VOLTS_PER_OCTAVE],
                default=MOOG_VOLTS_PER_OCTAVE,
            ),

            # RTC implementation
            # by default there is no RTC
            configuration.choice(
                "RTC_IMPLEMENTATION",
                choices=[
                    RTC_NONE,
                    RTC_DS3231,
                    RTC_DS1307,
                    RTC_NTP,
                ],
                default=RTC_NONE,
            ),

            # RTC Timezone offset for local time
            configuration.integer(
                "UTC_OFFSET_HOURS",
                minimum=-24,
                maximum=24,
                default=0,
            ),
            configuration.integer(
                "UTC_OFFSET_MINUTES",
                minimum=-59,
                maximum=59,
                default=0,
            ),

            # Wifi connection options
            # only applicable with Pico W, Pico 2 W
            configuration.choice(
                "WIFI_MODE",
                choices=[
                    WIFI_MODE_CLIENT,
                    WIFI_MODE_AP
                ],
                default=WIFI_MODE_AP,
            ),
            configuration.string(
                "WIFI_SSID",
                default="EuroPi",
            ),
            configuration.string(
                "WIFI_PASSWORD",
                default="europi",
            ),
            configuration.integer(
                "WIFI_CHANNEL",
                minimum=1,
                maximum=13,
                default=10,
            ),
            configuration.string(
                "WIFI_BSSID",
                default="",
            ),
            configuration.boolean(
                "WIFI_DHCP",
                default=True
            ),
            configuration.string(
                "WIFI_IP",
                default="192.168.4.1"
            ),
            configuration.string(
                "WIFI_NETMASK",
                default="255.255.255.0"
            ),
            configuration.string(
                "WIFI_GATEWAY",
                default="0.0.0.0"
            ),
            configuration.string(
                "WIFI_DNS",
                default="8.8.8.8"
            ),

            # WebREPL support
            # requires wifi configuration (above)
            configuration.boolean(
                "ENABLE_WEBREPL",
                default=False
            ),
        ]
        # fmt: on


def load_experimental_config():
    """
    Load the config points from /config/ExperimentalConfig.json

    If that file does not exist, or if it has missing keys, the default
    values are used to fill in any holes.
    """
    return ConfigFile.load_config(
        ExperimentalConfig, ConfigSpec(ExperimentalConfig.config_points())
    )



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A collection of classes and functions for dealing with configuration data. There are two main
types of configuration data: the `dict` configuration itself, and the `ConfigSpec`.

The configuration is represented by `Dict(str, Any)`, where the string key represents the config
point's name, pointing to a value that may be of any type.

The `ConfgSpec` is a collection of `ConfigPoints`. Each ConfigPoint consists of a name and description
of the valid values that it may have. There are several different types of COnfigPoints available.
"""

import os
import json
from file_utils import load_file, delete_file, load_json_file
from collections import namedtuple

Validation = namedtuple("Validation", "is_valid message")
"""A class containing configuration validation results.

    :param is_valid: True if the validation was successful, False otherwise.
    :param message: If `is_valid` is false this filed will contain a explanation message.
"""

VALID = Validation(is_valid=True, message="Valid.")
"""The default successful validation."""


class ConfigPoint:
    """Base class for defining `ConfigPoint` types.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param type: The name of this ConfigPoint's type
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(self, name: str, type: str, default, danger: bool = False):
        self.name = name
        self.type = type
        self.default = default
        self.danger = danger

    def validate(self, value) -> Validation:
        """Validates the given value with this ConfigPoint. Returns a `Validation` containing the
        validation result, as well as an error message containing the reason for a validation failure.
        """
        raise NotImplementedError


class StringConfigPoint(ConfigPoint):
    """A `ConfigPoint` that allows the input of arbitrary strings

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param type: The name of this ConfigPoint's type
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(self, name: str, default: str, danger: bool = False):
        super().__init__(name=name, type=float, default=default, danger=danger)

    def validate(self, value) -> Validation:
        """Validates the given value with this ConfigPoint. Returns a `Validation` containing the
        validation result, as well as an error message containing the reason for a validation failure.
        """
        if type(value) is str:
            return VALID
        else:
            return Validation(
                is_valid=False, message=f"Value {value} is of type {type(value)}, not str"
            )


class FloatConfigPoint(ConfigPoint):
    """A `ConfigPoint` that requires the selection from a range of floats.  The default value
    must lie within the specified range

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param minimum: The minimum allowed value
    :param maximum: The maximum allowed value
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(
        self, name: str, minimum: float, maximum: float, default: float, danger: bool = False
    ):
        super().__init__(name=name, type=float, default=default, danger=danger)
        self.maximum = maximum
        self.minimum = minimum

        if default < minimum or default > maximum:
            raise Exception(f"Defaulf {default} is out of range")

    def validate(self, value) -> Validation:
        """Validates the given value with this ConfigPoint. Returns a `Validation` containing the
        validation result, as well as an error message containing the reason for a validation failure.
        """
        # Silently caste integers into floats, so we don't crash if someone writes e.g. "0" instead of "0.0"
        if type(value) is int:
            value = float(value)
        if type(value) is float:
            if value >= self.minimum and value <= self.maximum:
                return VALID
            else:
                return Validation(is_valid=False, message=f"Value {value} is out of range")
        else:
            return Validation(is_valid=False, message=f"Value {value} is not a number")


class IntegerConfigPoint(ConfigPoint):
    """A `ConfigPoint` that requires selection from a range of integers. The default value
    must lie within the specified range

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param minimum: The minimum allowed value
    :param maximum: The maximum allowed value
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(self, name: str, minimum: int, maximum: int, default: int, danger: bool = False):
        super().__init__(name=name, type=int, default=default, danger=danger)
        self.minimum = minimum
        self.maximum = maximum

        if default < self.minimum or default > self.maximum:
            raise Exception(f"Default {default} is out of range")

    def validate(self, value) -> Validation:
        """Validates the given value with this ConfigPoint. Returns a `Validation` containing the
        validation result, as well as an error message containing the reason for a validation failure.
        """
        if type(value) is int:
            if value >= self.minimum and value <= self.maximum:
                return VALID
            else:
                return Validation(is_valid=False, message=f"Value {value} is out of range")
        else:
            return Validation(is_valid=False, message=f"Value {value} is not an integer")


class ChoiceConfigPoint(ConfigPoint):
    """A `ConfigPoint` that requires selection from a limited number of choices. The default value
    must exist in the given choices.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param choices: A List of all of the valid choices for this ConfigPoint
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(self, name: str, choices: list, default, danger: bool = False):
        if default not in choices:
            raise ValueError("default value must be available in given choices")
        super().__init__(name=name, type="choice", default=default, danger=danger)
        self.choices = choices

    def validate(self, value) -> Validation:
        result = value in self.choices
        if not result:
            return Validation(
                is_valid=result,
                message=f"Value '{value}' does not exist in valid choices: [{', '.join(str(c) for c in self.choices)}]",
            )
        return VALID


class BooleanConfigPoint(ChoiceConfigPoint):
    """A `ConfigPoint` that allows True/False values.

    :param name: The name of this `ConfigPoint`, will be used by scripts to look up the configured value.
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """

    def __init__(self, name: str, default: bool, danger: bool = False):
        super().__init__(name=name, choices=[False, True], default=default, danger=danger)


def boolean(name: str, default: bool, danger: bool = False) -> BooleanConfigPoint:
    """A helper function to simplify the creation of BooleanConfigPoints.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param default: The default value
    """
    return BooleanConfigPoint(name=name, default=default, danger=danger)


def choice(name: str, choices: list, default, danger: bool = False) -> ChoiceConfigPoint:
    """A helper function to simplify the creation of ChoiceConfigPoints. Requires selection from a
    limited number of choices. The default value must exist in the given choices.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param choices: A List of all of the valid choices for this ConfigPoint
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """
    return ChoiceConfigPoint(name=name, choices=choices, default=default, danger=danger)


def floatingPoint(
    name: str, minimum: float, maximum: float, default: float, danger: bool = False
) -> FloatConfigPoint:
    """A helper function to simplify the creation of FloatConfigPoints. Requires selection from a
    range of floats. The default value must exist in the given range.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param minumum: The minumum allowed value
    :param maximum: The maximum allowed value
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """
    return FloatConfigPoint(
        name=name, minimum=minimum, maximum=maximum, default=default, danger=danger
    )


def integer(
    name: str, minimum: int, maximum: int, default: int, danger: bool = False
) -> IntegerConfigPoint:
    """A helper function to simplify the creation of IntegerConfigPoints. Requires selection from a
    range of integers. The default value must exist in the given range.

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param minimum: The minimum allowed value
    :param maximum: The maximum allowed value
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """
    return IntegerConfigPoint(
        name=name, minimum=minimum, maximum=maximum, default=default, danger=danger
    )


def string(name: str, default: str, danger: bool = False) -> StringConfigPoint:
    """A helper function to simplify the creation of StringConfigPoints. Allows the input of
    any arbitrary string

    :param name: The name of this `ConfigPoint`, will be used by scripts to lookup the configured value.
    :param default: The default value
    :param danger: If true, mark this option as dangerous to modify in the config editor
    """
    return StringConfigPoint(name=name, default=default, danger=danger)


class ConfigSpec:
    """
    A container for `ConfigPoints` representing the set of configuration options for a specific
    script.
    """

    def __init__(self, config_points: list[ConfigPoint]) -> None:
        self.points = {}
        for point in config_points:
            if point.name in self.points:
                raise ValueError(f"config point {point.name} is already defined")
            self.points[point.name] = point

    def __len__(self):
        return len(self.points)

    def __iter__(self):
        return iter(self.points.values())

    def default_config(self) -> dict[str, object]:
        """Returns the default configuration for this spec."""
        return {point.name: point.default for point in self.points.values()}

    def validate(self, configuration) -> Validation:
        """Validates the given configuration with this spec. Returns a `Validation` containing the
        validation result, as well as an error message containing the reason for a validation failure.
        """
        for name, value in configuration.items():
            if name not in self.points:
                return Validation(is_valid=False, message=f"ConfigPoint '{name}' is not defined.")

            validation = self.points[name].validate(value)
            if not validation.is_valid:
                return validation

        return VALID


class ConfigFile:
    """A class containing functions for dealing with configuration files."""

    @staticmethod
    def load_from_file(path: str, config_spec: ConfigSpec):
        """
        Load the configuration settings from an arbitrary file

        :param path:  The path to the fole to read
        :param config_spec:  The specification of the configuration we're saving
        """
        if len(config_spec):
            saved_config = load_json_file(path)
            config = config_spec.default_config()
            validation = config_spec.validate(saved_config)

            if not validation.is_valid:
                raise ValueError(validation.message)

            config.update(saved_config)
            return ConfigSettings(config)
        else:
            return ConfigSettings({})

    @staticmethod
    def save_to_file(path: str, data: dict):
        """
        Save the configuration dict and save it to an arbitrary file

        Unlike save_config, this method will NOT create any directories; if you
        intend to save to a directory that may not exist, make sure to create it
        first.

        :param path:  The path to the file we're saving to
        :param dict:  The data to save
        """
        with open(path, "w") as file:
            # put newlines between items to make the resulting file easier to read
            # this makes debugging easier, in case human eyes are ever needed on the file
            json.dump(data, file, separators=(",\n", ":"))

    @staticmethod
    def config_filename(cls):
        """Returns the filename for the config file for the given class."""
        return f"config/{cls.__qualname__}.json"

    @staticmethod
    def save_config(cls, data: dict):
        """Take config as a dict and save to this class's config file.

        This will create the default configuration directory `/config` if it
        does not already exist

        .. note::
            Be mindful of how often `_save_config()` is called because
            writing to disk too often can slow down the performance of your
            script. Only call save state when state has changed and consider
            adding a time since last save check to reduce save frequency.
        """
        try:
            os.mkdir("config")
        except OSError:
            pass
        ConfigFile.save_to_file(ConfigFile.config_filename(cls), data)

    @staticmethod
    def load_config(cls, config_spec: ConfigSpec):
        """If this class has config points, this method validates and returns the ConfigSettings object
        representing the class's config file.  Otherwise an empty ConfigSettings object is returned.
        """
        return ConfigFile.load_from_file(ConfigFile.config_filename(cls), config_spec)

    @staticmethod
    def delete_config(cls):
        """Deletes the config file, effectively resetting to defaults."""
        delete_file(ConfigFile.config_filename(cls))


class ConfigSettings:
    """
    Collects the configuration settings into an object with attributes instead of a dict with keys

    :param d:  The raw dict loaded from the configuration file
    """

    def __init__(self, d):
        self.__dict__ = {}  # required for getattr & setattr
        self.__keys__ = set()

        for k in d.keys():
            self.validate_key(k)
            setattr(self, k, d[k])
            self.__keys__.add(k)

    def validate_key(self, key):
        """Ensures that a `dict` key is a valid attribute name

        :param key:  The string to check

        :return:     True if the key is valid. Otherwise an exception is raised

        :raises ValueError: if the key contains invalid characters; only letters, numbers, hyphens, and underscores
            are permitted. They key cannot be length 0, nor can it begin with a number
        """
        key = key.strip()
        for ch in key:
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                raise ValueError(
                    f"Invalid attribute name: {key}. Keys cannot contain the character {ch}"
                )

        if len(key) == 0:
            raise ValueError("Invalid attribute name: key cannot be empty")
        elif key[0].isdigit():
            raise ValueError("Invalid attribute name: key cannot start with a number")

        return True

    def keys(self):
        return self.__keys__

    def __getitem__(self, k):
        return getattr(self, k)

    def __eq__(self, that):
        """Allows comparing the config object directly to either another config object or a dict

        :param that:  The object we're comparing to, either a dict or another ConfigSettings object

        :return True: if the two objects are equivalent, otherwise False
        """
        if type(that) is dict:
            try:
                that = ConfigSettings(that)
                return self == that
            except ValueError:
                return False
        elif type(that) is ConfigSettings:
            return self.__dict__ == that.__dict__



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript
from time import sleep
from version import __version__


class About(EuroPiScript):
    """
    A simple EuroPiScript to display the EuroPi version

    Also displays the CPU frequency mode and Pico model
    """

    def __init__(self):
        super().__init__()

    def main(self):
        turn_off_all_cvs()

        oled.centre_text(
            f"""EuroPi v{__version__}
{europi_config.EUROPI_MODEL}/{europi_config.PICO_MODEL}
{europi_config.CPU_FREQ}"""
        )

        while True:
            sleep(1)


if __name__ == "__main__":
    About().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from experimental.clocks.clock_source import ExternalClockSource
from utime import ticks_ms


class NullClock(ExternalClockSource):
    """
    A placeholder replacement for a realtime clock.

    This class sets the base time to 1 January, 1970, 00:00:00.0 and adds the current value of
    ticks_ms() to that time. This allows the RTC to tick up, but won't actually be indicitave of the real time.

    Because ticks_ms() rolls over ever 2**30 ms, the datetime returned by this clock will also roll over
    at the same interval. BUT, 2**30 ms ~= 298 hours, so unless your EuroPi is powered-on 12 days continuously
    this won't ever cause problems.
    """

    def __init__(self):
        super().__init__()
        self.last_check_at = ticks_ms()

    def set_datetime(self, datetime):
        # we don't allow configuring the time here
        pass

    def datetime(self):
        t = ticks_ms()

        s = (t // 1000) % 60
        m = (t // (1000 * 60)) % 60
        h = (t // (1000 * 60 * 60)) % 24
        dd = (t // (1000 * 60 * 60 * 24)) % 31
        mm = 1
        yy = 1970
        wd = (4 + dd) % 7  # 1 jan 1970 was a thursday
        yd = 0  # ignore yearday

        return (yy, mm, dd, h, m, s, wd, yd)



"""This 'experimental' package is a place for code to live that might change significantly or might
not be fully thought out or feature rich enough to include in the firmware proper. It exists for the
purpose of allowing developers to experiment, colaborate, write scripts against it and provide
feedback. Code here should slowly evolve and stabilize.

Once code here has proven to be both stable and useful we can promote it out of the experimental
into the firmware library package and update calling scripts.

Code that is neither stable nor useful should be removed.
"""



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Additional classes and utilities for multi-threading support using _thread
"""

from europi import *


class DigitalInputHelper:
    """Helper class for reading digital inputs & outputs without using ISRs

    _thread can lock up if ISRs fire too frequently (e.g. pressing B1 repeatedly to select a mode). To avoid using
    ISRs this class can be used in your main thread to simulate the behaviour of ISRs by running a callback function
    when the state of the digital input changes.

    In your main loop you must call `.update()` to read the current state of the inputs & invoke the callback
    functions

    Calling `.update()` will populate the following fields of the DigitalInputHelper:
    * b1_last_pressed: the ms tick time that b1 was last pressed
    * b1_last_released: the ms tick time that b1 was last released
    * b1_pressed: True if b1 is currently pressed down, otherwise False
    * b1_rising: True if b1 transitioned to the pressed state during the last update() call, otherwise False
    * b1_falling: True if b1 transitioned to the released state during the last update() call, otherwise False
    * b2_last_pressed: the ms tick time that b2 was last pressed
    * b2_last_released: the ms tick time that b2 was last released
    * b2_pressed: True if b2 is currently pressed down, otherwise False
    * b2_rising: True if b1 transitioned to the pressed state during the last update() call, otherwise False
    * b2_falling: True if b1 transitioned to the released state during the last update() call, otherwise False
    * din_last_rise: the ms tick time that din last received a rising signal
    * din_last_fall: the ms tick time that din last received a falling signal
    * din_high: True if din is currently high, otherwise False
    * din_rising: True if din transitioned to the high state during the last update() call, otherwise False
    * din_falling: True if b1 transitioned to the low state during the last update() call, otherwise False

    :param on_din_rising:   Callback function to invoke when din detects a rising edge
    :param on_din_falling:  Callback function to invoke when din detects a falling edge
    :param on_b1_rising:    Callback function to invoke when b1 detects a rising edge
    :param on_b1_falling:   Callback function to invoke when b1 detects a falling edge
    :param on_b2_rising:    Callback function to invoke when b2 detects a rising edge
    :param on_b2_falling:   Callback function to invoke when b2 detects a falling edge
    """

    def __init__(
        self,
        on_din_rising=lambda: None,
        on_din_falling=lambda: None,
        on_b1_rising=lambda: None,
        on_b1_falling=lambda: None,
        on_b2_rising=lambda: None,
        on_b2_falling=lambda: None,
    ):
        # Connect the callback functions
        self.on_din_rising = on_din_rising
        self.on_din_falling = on_din_falling
        self.on_b1_rising = on_b1_rising
        self.on_b1_falling = on_b1_falling
        self.on_b2_rising = on_b2_rising
        self.on_b2_falling = on_b2_falling

        # Rising/falling states; these get updated inside the `update()` function
        self.din_rising = False
        self.din_falling = False
        self.b1_rising = False
        self.b1_falling = False
        self.b2_rising = False
        self.b2_falling = False

        # High/low flags for each input
        self.din_high = False
        self.b1_pressed = False
        self.b2_pressed = False

        # MS ticks when events were last recorded
        self.b1_last_pressed = 0
        self.b1_last_released = 0
        self.b2_last_pressed = 0
        self.b2_last_released = 0
        self.din_last_rise = 0
        self.din_last_fall = 0

    def update(self):
        """Check the state of the digital inputs, call the callback functions when appropriate

        This function should be called inside the main loop of your program
        """
        din_state = din.value() != 0
        b1_state = b1.value() != 0
        b2_state = b2.value() != 0

        self.din_rising = not self.din_high and din_state
        self.din_falling = self.din_high and not din_state

        self.b1_rising = not self.b1_pressed and b1_state
        self.b1_falling = self.b1_pressed and not b1_state

        self.b2_rising = not self.b2_pressed and b2_state
        self.b2_falling = self.b2_pressed and not b2_state

        self.din_high = din_state
        self.b1_pressed = b1_state
        self.b2_pressed = b2_state

        if self.b1_rising:
            self.b1_last_pressed = time.ticks_ms()
            self.on_b1_rising()
        elif self.b1_falling:
            self.b1_last_released = time.ticks_ms()
            self.on_b1_falling()

        if self.b2_rising:
            self.b2_last_pressed = time.ticks_ms()
            self.on_b2_rising()
        elif self.b2_falling:
            self.b2_last_released = time.ticks_ms()
            self.on_b2_falling()

        if self.din_rising:
            self.din_last_rise = time.ticks_ms()
            self.on_din_rising()
        elif self.din_falling:
            self.din_last_fall = time.ticks_ms()
            self.on_din_falling()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License."""A simple screensaver class for the EuroPi
"""
A simple screensaver implementation for EuroPi

@author Chris Iverach-Brereton <ve4cib@gmail.com>
@year   2023
"""

import random
import utime

from europi import *


class Screensaver:
    """A screensaver for the EuroPi

    This class' draw() function should be called from inside the application's main loop
    to draw the screensaver.

    At regular intervals the EuroPi logo will be drawn to a random location on the screen.

    Alternatively draw_blank() will simply clear the screen
    """

    ## The EuroPi logo as a 16x16 image
    LOGO = bytearray(
        b"\x01\xc0\x03p;\x18n1\xc0c\xf8\xc6\x0c\x93\r\xb9\x19-\xf37\xc6\x12\x0c\x98\x19\xc83L\x1el\x0c8"
    )
    LOGO_WIDTH = 16
    LOGO_HEIGHT = 16

    ## Standard duration before we activate the screensaver
    ACTIVATE_TIMEOUT_MS = 1000 * 60 * 5
    ACTIVATE_TIMEOUT_US = ACTIVATE_TIMEOUT_MS * 1000

    ## Standard duration before we blank the screen
    BLANK_TIMEOUT_MS = 1000 * 60 * 20
    BLANK_TIMEOUT_US = BLANK_TIMEOUT_MS * 1000

    def __init__(self):
        self.last_logo_reposition_at = 0

    def draw(self, force=False):
        """Draw the logo to a random position on the screen

        If the current ticks_ms is an even multiple of the screensaver interval,
        or the `force` paramter is True, we will draw.  Otherwise this does
        nothing

        :param force:  Force the logo to be drawn, regardless of the current ticks
        """
        LOGO_UPDATE_INTERVAL = 2000

        now = utime.ticks_ms()
        elapsed_ms = time.ticks_diff(now, self.last_logo_reposition_at)
        if force or abs(elapsed_ms) >= LOGO_UPDATE_INTERVAL:
            self.last_logo_reposition_at = now
            x = random.randint(0, OLED_WIDTH - self.LOGO_WIDTH)
            y = random.randint(0, OLED_HEIGHT - self.LOGO_HEIGHT)

            oled.fill(0)
            fb = FrameBuffer(self.LOGO, self.LOGO_WIDTH, self.LOGO_HEIGHT, MONO_HLSB)
            oled.blit(fb, x, y)
            oled.show()

    def draw_blank(self):
        """Blank the screen completely"""
        oled.fill(0)
        oled.show()


class OledWithScreensaver:
    """A wrapper for europi.oled that provides automatic screensaver & blank if enabled

    The notify_user_interaction() method should be called in the ISR for any events that should interrupt or
    defer the screensaver/blank

    Set .enable_screensaver or .enable_blank to False to disable the screensaver/blanking activation completely
    (though if you do that, just use europi.oled instead of this class)

    :param enable_screensaver:  If true, the screensaver will activate when needed
    :param enable_blank:        If true, the screen will blank after the screensaver has been active for a while
    """

    def __init__(self, enable_screensaver=True, enable_blank=True):
        self.screensaver = Screensaver()
        self.enable_screensaver = enable_screensaver
        self.enable_blank = enable_blank

        self.show_screensaver = False
        self.show_blank = False

        self.last_user_interaction_at = utime.ticks_ms()

    def is_screenaver(self):
        """Is the screensaver currently showing?"""
        return self.show_screensaver

    def is_blank(self):
        """Is the screen blanked due to inactivity?"""
        return self.show_blank

    def notify_user_interaction(self):
        """Notifies the screensaver subsystem that the user has physically interacted with the module
        and that the screensaver should defer/shutdown as appropriate
        """
        self.last_user_interaction_at = utime.ticks_ms()

    def show(self):
        now = utime.ticks_ms()
        if (
            self.enable_blank
            and utime.ticks_diff(now, self.last_user_interaction_at)
            > self.screensaver.BLANK_TIMEOUT_MS
        ):
            self.show_blank = True
            self.show_screensaver = False
            self.screensaver.draw_blank()
        elif (
            self.enable_screensaver
            and utime.ticks_diff(now, self.last_user_interaction_at)
            > self.screensaver.ACTIVATE_TIMEOUT_MS
        ):
            self.show_screensaver = True
            self.show_blank = False
            self.screensaver.draw()
        else:
            self.show_blank = False
            self.show_screensaver = False
            oled.show()

    # The following are just wrappers for the functions in the Display class to allow 1:1 access
    # See europi.Display for documentation details

    def fill(self, color):
        oled.fill(color)

    def text(self, string, x, y, color=1):
        oled.text(string, x, y, color)

    def centre_text(self, text, clear_first=True, auto_show=True):
        oled.centre_text(text, clear_first=clear_first, auto_show=False)
        if auto_show:
            self.show()

    def line(self, x1, y1, x2, y2, color=1):
        oled.line(x1, y1, x2, y2, color)

    def hline(self, x, y, length, color=1):
        oled.hline(x, y, length, color)

    def vline(self, x, y, length, color=1):
        oled.vline(x, y, length, color)

    def rect(self, x, y, width, height, color=1):
        oled.rect(x, y, width, height, color)

    def fill_rect(self, x, y, width, height, color=1):
        oled.fill_rect(x, y, width, height, color)

    def ellipse(self, x, y, xr, yr, colour=1, fill=False):
        oled.ellipse(x, y, xr, yr, colour, fill)

    def blit(self, buffer, x, y):
        oled.blit(buffer, x, y)

    def scroll(self, x, y):
        oled.scroll(x, y)

    def invert(self, color=1):
        oled.invert(color)

    def contrast(self, contrast):
        oled.contrast(contrast)

    def pixel(self, x, y, color=1):
        oled.pixel(x, y, color)



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A simple HTTP server for the Raspberry Pi Pico
"""

import json

try:
    import socket
except ImportError as err:
    raise Exception(f"Failed to load HTTP server dependencies: {err}")

from europi_log import *


class HttpStatus:
    """
    HTTP status codes to include in the header.

    This collection is not exhaustive, just what we need to handle this minimal server implementation
    """

    # 200 series - everything's fine
    OK = 200

    # 400 series - error is on the client end
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_FOUND = 404
    REQUEST_TIMEOUT = 408
    TEAPOT = 418

    # 500 series - error is on the server end
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501

    # Human-readable names/descriptions of the error codes above
    # If you add another error code, make sure to add it here too!
    StatusText = {
        OK: "OK",
        BAD_REQUEST: "Bad Request",
        UNAUTHORIZED: "Unauthorized",
        FORBIDDEN: "Forbidden",
        NOT_FOUND: "Not Found",
        REQUEST_TIMEOUT: "Request Timeout",
        TEAPOT: "I'm a teapot",
        INTERNAL_SERVER_ERROR: "Internal Server Error",
        NOT_IMPLEMENTED: "Not Implemented",
    }


class MimeTypes:
    """
    Common MIME types we can support with this HTTP server implementation
    """

    CSV = "text/csv"
    HTML = "text/html"
    JSON = "text/json"
    TEXT = "text/plain"
    XML = "text/xml"
    YAML = "text/yaml"


class HttpServer:
    """
    A basic HTTP server for EuroPi.

    This class will open a socket on the specified port, allowing for clients to connect to us.
    Only HTTP is supported, not HTTPS. Basic GET/POST requests should work, but websockets
    and anything fancy likely won't.

    Port 80 is officially reserved for HTTP traffic, and can be used by default. Port 8080
    is also commonly used for HTTP traffic.

    If you're operating in WiFi Client mode you may be subject to the whims of whatever
    port filter/firewall your IT administrator has implemented, so some ports may not work/
    may not be available.

    Operating in WiFi AP mode should allow the use of any port you want.

    You should define a callback to handle incoming GET and/or POST requests, e.g.:

    .. code-block:: python

        server = HttpServer(port=8080)

        @server.get_handler
        def handle_http_get(request:str=None, connection:Socket=None):
            # process the request
            server.send_response(...)

        @server.post_handler
        def handle_http_post(request:str=None, connection:Socket=None):
            # process the request
            server.send_response(...)

    Responses can be an HTTP page, plain text, or JSON/CSV/YAML/XML formatted data. See
    MimeTypes for supported types. The response should always be a string; if sending
    a dict as JSON data you'll need to stringify it before passing it to send_response.

    You may send your own error codes as desired:

    .. code-block:: python

        def handle_http_request(connection=None, request=None):
            srv.send_error_page(
                Exception("We're out of coffee!")
                connection,
                HttpStatus.TEAPOT,  # send error 418 "I'm a teapot"
            )

    Inside the program's main loop you should call srv.check_connections() to process any
    incoming requests:

    .. code-block:: python

        def main(self):
            # ...
            while True:
                # ...
                srv.check_requests()
                # ...

    :param port:  The port to listen on
    """

    # A basic error page template
    # Contains 3 parameters for the user to fill:
    # - errno: the error number (e.g. 404, 500)
    # - errname: the human-legible error name (e.g Not Found, Internal Server Error)
    # - message: free-form error message, stack trace, etc...
    ERROR_PAGE = """<!DOCTYPE html>
<!DOCTYPE html>
<html lang="en">
    <head>
        <style>
            body {{
                font-family: Montserrat;
                text-align: center;
            }}
            h1 {{
                font-weight: normal;
                font-size: 2.5rem;
                letter-spacing: 1.75rem;
                padding-left: 1.75rem;
            }}
            h2 {{
                font-weight:  bold;
                font-size: 3.0rem;
            }}
            p {{
                font-weight: lighter;
                font-size: 2.0rem;
            }}
            .content-wrapper {{
                margin: 0;
                position: absolute;
                top: 50%;
                align-content: center;
                width: 100%;
                -ms-transform: translateY(-50%);
                transform: translateY(-50%);
            }}
        </style>
        <title>EuroPi Error {errno}</title>
    </head>
    <body>
        <div class="content-wrapper"?>
            <h1>EuroPi</h1>
            <h2>{errno} {errname}</h2>
            <p>
                {message}
            </p>
        </div>
    </body>
</html>
"""

    def __init__(self, port=80):
        self.port = port
        self.get_callback = self.default_request_handler
        self.post_callback = self.default_request_handler

        self.socket = socket.socket()
        addr = socket.getaddrinfo("0.0.0.0", port)
        addr = addr[0][-1]
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.settimeout(0)
        self.socket.bind(addr)
        self.socket.listen(5)

    def default_request_handler(self, connection=None, request=None):
        """
        The default request handler for GET and POST requests the server.

        The intent is that whatever program is using the HTTP server will create their own callback
        to replace this function. So all we do is raise a NotImplementedError that's handled
        by self.check_requests() and will serve our HTTP 501 error page accordingly.

        :param connection:  The socket the client connected on
        :param request:  The client's request

        :raises NotImplementedError: This results in an HTTP 501 error response
        """
        raise NotImplementedError("No request handler set")

    def check_requests(self):
        """
        Poll the socket and process any incoming requests.

        This will invoke the request handler function if it has been defined

        This function should be called inside the main loop of the program

        The client socket is closed after we send our response
        """
        conn = None
        while True:
            try:
                conn, addr = self.socket.accept()

                # the request should look something like this
                #
                # GET / HTTP/1.1
                # Host: localhost:8080
                # User-Agent: Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:135.0) Gecko/20100101 Firefox/135.0
                # Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
                # Accept-Language: en-US,en;q=0.5
                # Accept-Encoding: gzip, deflate, br, zstd
                # Connection: keep-alive
                # Upgrade-Insecure-Requests: 1
                # Sec-Fetch-Dest: document
                # Sec-Fetch-Mode: navigate
                # Sec-Fetch-Site: none
                # Sec-Fetch-User: ?1
                # Priority: u=0, i

                request = conn.recv(1024)
                request = request.decode("UTF-8")

                if request.startswith("GET"):
                    self.get_callback(request=request, connection=conn)
                elif request.startswith("POST"):
                    self.post_callback(request=request, connection=conn)
                else:
                    method = request.strip().split()[0]
                    self.send_error_page(
                        Exception(f"Unsupported HTTP method {method}"),
                        conn,
                        status=HttpStatus.BAD_REQUEST,
                        headers=None,
                    )
            except NotImplementedError as err:
                log_warning(f"{err}", "http_server")
                # send a 501 error page
                self.send_error_page(err, conn, HttpStatus.NOT_IMPLEMENTED)
            except OSError as err:
                return
            except Exception as err:
                log_warning(f"{err}", "http_server")
                # send a 500 error page
                self.send_error_page(err, conn, HttpStatus.INTERNAL_SERVER_ERROR)
            finally:
                if conn is not None:
                    conn.close()
                    conn = None

    def get_handler(self, func):
        """
        Decorator for the function to handle HTTP GET requests

        The provided function must accept the following keyword arguments:
        - request: str  The request the client sent
        - conn: socket  A socket connection to the client

        :param func:  The function to handle the request.
        """

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        self.get_callback = wrapper
        return wrapper

    def post_handler(self, func):
        """
        Decorator for the function to handle HTTP POST requests

        The provided function must accept the following keyword arguments:
        - request: str  The request the client sent
        - conn: socket  A socket connection to the client

        :param func:  The function to handle the request.
        """

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        self.post_callback = wrapper
        return wrapper

    def send_error_page(
        self,
        error,
        connection,
        status=HttpStatus.INTERNAL_SERVER_ERROR,
        headers=None,
    ):
        """
        Serve our customized HTTP error page

        :param error:  The exception that caused the error
        :param connection:  The socket to send the response over
        :param status:  The error status to respond with
        :param headers:  Optional additional headers
        """
        self.send_html(
            connection,
            self.ERROR_PAGE.format(
                errno=status,
                errname=HttpStatus.StatusText[status],
                message=str(error),
            ),
            status=status,
            headers=headers,
        )

    def send_json(self, connection, data, headers=None):
        """
        Send a JSON object to the client

        :param connection:  The socket connection to the client
        :param data:  A dict to be converted to a JSON object
        :param headers:  Optional additional HTTP headers to include
        """
        self.send_response(
            connection,
            json.dumps(data),
            status=HttpStatus.OK,
            content_type=MimeTypes.JSON,
            headers=headers,
        )

    def send_html(self, connection, html_page, status=HttpStatus.OK, headers=None):
        """
        Send an HTML document to the client

        :param connection:  The socket to send the data over
        :param html_page:  A string containing the HTML document.
        :param status:  The HTTP status to send the page with
        :param headers:  Optional additional HTTP headers
        """
        self.send_response(
            connection, html_page, content_type=MimeTypes.HTML, status=status, headers=headers
        )

    def send_response(
        self,
        connection,
        response,
        status=HttpStatus.OK,
        content_type=MimeTypes.HTML,
        headers=None,
    ):
        """
        Send a response to the client

        :param connection:  The socket connection to the client
        :param response:  The response payload
        :param status:  The HTTP status to respond with
        :param content_type:  The MIME type to include in the HTTP header
        :param headers:  Optional dict of key/value pairs for addtional HTTP headers. Charset is ALWAYS utf-8
        """
        header = f"HTTP/1.0 {status} {HttpStatus.StatusText[status]}\r\nContent-type: {content_type}\r\ncharset=utf-8"

        if headers is not None:
            for k in headers.keys():
                header = f"{header}\r\n{k}={headers[k]}"

        connection.send(f"{header}\r\n\r\n".encode("UTF-8"))
        connection.send(response.encode("UTF-8"))



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import framebuf

from machine import I2C
from machine import Pin
from ssd1306 import SSD1306_I2C

from europi import (
    europi_config,
    TEST_ENV,
    CHAR_HEIGHT,
)
from europi_display import Display as BasicDisplay


# TODO: add a method to select the font to use by default


class CustomFontWriter:
    def __init__(self, device, font):
        """Initialize the Writer.

        device: the OLED display instance
        font: a font module
        """
        self.device = device
        self.font = font
        if font.hmap():
            self.map = framebuf.MONO_HMSB if font.reverse() else framebuf.MONO_HLSB
        else:
            raise ValueError("Font must be horizontally mapped.")
        self.screenwidth = device.width  # In pixels
        self.screenheight = device.height

    def print(self, string, x, y, c=1):
        """Print the string using the x, y coordinates as the upper-left corner of the text.
        With c=0, the text is display as black text on white background.
        """
        for char in string:
            if char == "\n":  # line breaks are ignored
                return
            glyph, char_height, char_width = self.font.get_ch(char)
            buf = bytearray(glyph)
            if c != 1:
                for i, v in enumerate(buf):
                    buf[i] = 0xFF & ~v
            fbc = framebuf.FrameBuffer(buf, char_width, char_height, self.map)
            self.device.blit(fbc, x, y)
            x += char_width

    def string_len(self, string):
        """Returns the length of string in pixels."""
        n = 0
        for char in string:
            n += self._char_len(char)
        return n

    def _char_len(self, char):
        """Returns the length of char in pixels."""
        if char == "\n":
            char_width = 0
        else:
            _, _, char_width = self.font.get_ch(char)
        return char_width


class CustomFontDisplay(BasicDisplay):
    """A class for adding custom font capability to the OLED display.

    This class can be used identically to the parent class, with the added
    parameter of 'font' which can be passed when using any of the methods to
    specify an alternative to the default 8x8 pixel font
    """

    def __init__(self, default_font=None):  # by default will use the monospaced 8x8 font
        self.writers = {}  # re-usable large font writer instances
        self.default_font = default_font
        super().__init__(
            width=europi_config.DISPLAY_WIDTH,
            height=europi_config.DISPLAY_HEIGHT,
            sda=europi_config.DISPLAY_SDA,
            scl=europi_config.DISPLAY_SCL,
            channel=europi_config.DISPLAY_CHANNEL,
            freq=europi_config.DISPLAY_FREQUENCY,
            contrast=europi_config.DISPLAY_CONTRAST,
            rotate=europi_config.ROTATE_DISPLAY,
        )

    def _writer(self, font):
        """Returns the large font writer for the specified font."""
        n = font.__name__
        if n not in self.writers:
            self.writers[n] = CustomFontWriter(self, font)
        return self.writers[n]

    def text_width(self, s, font=None):
        """Returns the length of the string s in pixels."""
        if font or self.default_font:
            return self._writer(font or self.default_font).string_len(s)
        else:
            return len(s) * CHAR_WIDTH

    def text_height(self, s=None, font=None):
        """Returns the height of the font s in pixels.
        The name text_height() is to be coherent with the text_width() method above.
        The string s is optional and ignored if present. The parameter is
        there for a future version which could return a more precise value based on
        the supplied text.
        """
        if font:
            return font.height()
        elif self.default_font:
            return self.default_font.height()
        else:
            return CHAR_HEIGHT

    def text(self, s, x, y, c=1, font=None):
        """Display the string s using the x, y coordinates as the upper-left corner of the text."""
        if font or self.default_font:
            self._writer(font or self.default_font).print(s, x, y, c)
        else:
            super().text(s, x, y, c)

    def centre_text(self, text, clear_first=True, auto_show=True, font=None):
        """Split the provided text across 3 lines of display."""

        # Default font is 8x8 pixel monospaced font which can be split to a
        # maximum of 4 lines on a 128x32 display, but we limit it to 3 lines
        # for readability.

        if font or self.default_font:
            # f = font or self.default_font

            if clear_first:
                self.fill(0)

            lines = str(text).split("\n")
            maximum_lines = self.height // self.text_height(font=font)
            if len(lines) > maximum_lines:
                raise Exception("Provided text exceeds available space on oled display.")

            h = self.text_height(font=font) + 1
            padding_top = (self.height - (len(lines) * h)) // 2

            for index, content in enumerate(lines):
                w = self.text_width(content, font=font)
                x_offset = int((self.width - w) / 2) - 1
                y_offset = index * h + padding_top
                self.text(content, x_offset, y_offset, font=font)

        else:
            super().centre_text(text, clear_first=clear_first, auto_show=auto_show)

        if auto_show:
            self.show()


oled = CustomFontDisplay()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Classes and definitions for interacting with the OLED display
"""

from machine import I2C, Pin
import ssd1306
from ssd1306 import SSD1306_I2C

# Default font is 8x8 pixel monospaced font.
CHAR_WIDTH = 8
CHAR_HEIGHT = 8


class Display(SSD1306_I2C):
    """
    A class for drawing graphics and text to the OLED.

    The OLED Display works by collecting all the applied commands and only
    updates the physical display when ``oled.show()`` is called. This allows
    you to perform more complicated graphics without slowing your program, or
    to perform the calculations for other functions, but only update the
    display every few steps to prevent lag.

    To clear the display, simply fill the display with the colour black by using ``oled.fill(0)``

    More explanations and tips about the the display can be found in the oled_tips file
    `oled_tips.md <https://github.com/Allen-Synthesis/EuroPi/blob/main/software/oled_tips.md>`_
    """

    def __init__(
        # fmt: off
        self,
        width,
        height,
        sda,
        scl,
        channel,
        freq,
        contrast,
        rotate
        # fmt: on
    ):
        i2c = I2C(channel, sda=Pin(sda), scl=Pin(scl), freq=freq)
        self.width = width
        self.height = height
        super().__init__(self.width, self.height, i2c)
        self.rotate(rotate)
        self.contrast(contrast)

    def rotate(self, rotate):
        """Flip the screen from its default orientation

        :param rotate:  True or False, indicating whether we want to flip the screen from its default orientation
        """
        # From a hardware perspective, the default screen orientation of the display _is_ rotated
        # But logically we treat this as right-way-up.
        if rotate:
            rotate = 0
        else:
            rotate = 1
        self.write_cmd(ssd1306.SET_COM_OUT_DIR | ((rotate & 1) << 3))
        self.write_cmd(ssd1306.SET_SEG_REMAP | (rotate & 1))

    def centre_text(self, text, clear_first=True, auto_show=True):
        """Display one or more lines of text centred both horizontally and vertically.

        :param text:  The text to display
        :param clear_first:  If true, the screen buffer is cleared before rendering the text
        :param auto_show:  If true, oled.show() is called after rendering the text. If false, you must call oled.show() yourself
        """
        if clear_first:
            self.fill(0)
        # Default font is 8x8 pixel monospaced font which can be split to a
        # maximum of 4 lines on a 128x32 display, but the maximum_lines variable
        # is rounded down for readability
        lines = str(text).split("\n")
        maximum_lines = round(self.height / CHAR_HEIGHT)
        if len(lines) > maximum_lines:
            raise Exception("Provided text exceeds available space on oled display.")
        padding_top = (self.height - (len(lines) * (CHAR_HEIGHT + 1))) / 2
        for index, content in enumerate(lines):
            x_offset = int((self.width - ((len(content) + 1) * (CHAR_WIDTH - 1))) / 2) - 1
            y_offset = int((index * (CHAR_HEIGHT + 1)) + padding_top) - 1
            self.text(content, x_offset, y_offset)

        if auto_show:
            self.show()


class DummyDisplay:
    """
    An alternative to Display that provides software compatibility, but no hardware interface

    Useful e.g when debugging a breadboard version of the module without a display connected
    """

    def __init__(self, width, height):
        self.width = width
        self.height = height

    def rotate(self, rotate):
        pass

    def centre_text(self, text, clear_first=True, auto_show=True):
        pass

    def show(self):
        pass

    def fill(self, color):
        pass

    def text(self, string, x, y, color=1):
        pass

    def line(self, x1, y1, x2, y2, color=1):
        pass

    def hline(self, x, y, length, color=1):
        pass

    def vline(self, x, y, length, color=1):
        pass

    def rect(self, x, y, width, height, color=1):
        pass

    def fill_rect(self, x, y, width, height, color=1):
        pass

    def ellipse(self, x, y, xr, yr, colour=1, fill=False):
        pass

    def blit(self, buffer, x, y):
        pass

    def scroll(self, x, y):
        pass

    def invert(self, color=1):
        pass

    def contrast(self, contrast):
        pass

    def pixel(self, x, y, color=1):
        pass



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Assorted mathematical and statistical functions that can be re-used across scripts

Intended to augment Python's standard math library with additional useful functions
"""


def prod(l):
    """
    Calculate the product of all items in a list

    Equivalent the Python3's math.prod

    :param l:  An iterable collection of numbers

    :return:  The product of all items in the list, or 0 if the list is empty
    """
    if len(l) == 0:
        return 0
    p = 1
    for n in l:
        p = p * n
    return p


def median(l):
    """
    Calculate the median value from an array

    This is a cheater median, as we always choose the lower value if the length is even instead of
    averaging the middle two values. This is faster, but mathematically incorrect.

    :param l:  An iterable collection of numbers

    :return:  The median value from the list, or 0 if the list is empty
    """
    if len(l) == 0:
        return 0
    arr = [i for i in l]
    arr.sort()
    return arr[len(arr) // 2]


def mean(l):
    """
    Calculate the arithmetic mean value from an array

    :param l:  An iterable collection of numbers

    :return:  The arithmetic mean of the list, or 0 of the list is empty
    """
    if len(l) == 0:
        return 0
    return sum(l) / len(l)


def geometric_mean(l):
    """
    Calculate the geometric mean value of an array

    :param l:  An iterable collection of numbers

    :return:  The geometric mean of the list, or 0 of the list is empty
    """
    if len(l) == 0:
        return 0
    return prod(l) ** (1.0 / len(l))


def harmonic_mean(l):
    """
    Calculate the harmonic mean value of an array

    :param l:  An iterable collection of numbers

    :return:  The harmonic mean of the list, or 0 of the list is empty
    """
    if len(l) == 0:
        return 0
    s = 0
    for n in l:
        s += 1.0 / n
    return len(l) / s


def mode(l):
    """
    Calculate the mode of an array

    The mode is the most-occurring item; if multiple items are tied,
    the median of the tied items is returned

    :param l:  An iterable collection of numbers

    :return:  The mode of the list, or 0 if the list is empty
    """
    if len(l) == 0:
        return 0
    counts = {}
    for n in l:
        if n in counts:
            counts[n] += 1
        else:
            counts[n] = 1

    max_count = 0
    tied_for_max = []
    for n in l:
        if counts[n] > max_count:
            max_count = counts[n]
            tied_for_max = [n]
        elif counts[n] == max_count:
            tied_for_max.append(n)

    if len(tied_for_max) == 1:
        return tied_for_max[0]
    else:
        tied_for_max.sort()
        return tied_for_max[len(tied_for_max) // 2]


def rescale(x, old_min, old_max, new_min, new_max, clip=True):
    """
    Convert x in [old_min, old_max] -> y in [new_min, new_max] using linear interpolation

    :param x:  The value to convert
    :param old_min:  The old (inclusive) minimum
    :param old_max:  The old (inclusive) maximum
    :param new_min:  The new (inclusive) minimum
    :param new_max:  The new (inclusive) maximum
    :param clip:     If true, we clip values within the [min, max] range; otherwise we extrapolate based on the ranges
    """
    if clip and x < old_min:
        return new_min
    elif clip and x > old_max:
        return new_max
    else:
        return (x - old_min) * (new_max - new_min) / (old_max - old_min) + new_min


def solve_linear_system(m):
    """
    Use gaussian elimination to solve a series of linear equations

    The provided matrix is the augmented matrix representation:

    .. code-block::

        [
            [k11 k12 k12 ... k1n  a1]
            [k21 k22 k22 ... k2n  a2]
            .
            .
            .
            [kmn kmn kmn ... kmn  am]
        ]

    :param m:  The augmented matrix representation of the series of equations. This array is mangled in the process
              of calculation

    :return:   A matrix of the coefficients of the equation
    """
    n_eqs = len(m)

    # sort the rows
    for i in range(n_eqs):
        for j in range(i + 1, n_eqs):
            if abs(m[i][i]) < abs(m[j][i]):
                # swap rows i and j with each other
                for k in range(n_eqs + 1):
                    tmp = m[j][k]
                    m[j][k] = m[i][k]
                    m[i][k] = tmp

    # gaussian elimination
    for i in range(n_eqs - 1):
        for j in range(i + 1, n_eqs):
            f = m[j][i] / m[i][i]
            for k in range(n_eqs + 1):
                m[j][k] = m[j][k] - f * m[i][k]

    # back substitution
    results = list(range(n_eqs))
    for i in range(n_eqs - 1, -1, -1):
        results[i] = m[i][n_eqs]
        for j in range(i + 1, n_eqs):
            if i != j:
                results[i] = results[i] - m[i][j] * results[j]
        results[i] = results[i] / m[i][i]

    return results


def gray_encode(n: int) -> int:
    """
    Convert a binary integer to its Gray Encoding equivalent.

    :param n:  The value to convert
    """
    return (n & 0xFF) ^ ((n & 0xFF) >> 1)


def gray_decode(n: int) -> int:
    """
    Convert a binary integer from Gray Encoding to its traditional equivalent.

    :param n:  The value to convert
    """
    mask = n
    while mask != 0:
        mask = mask >> 1
        n = n ^ mask
    return n



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Additional random functions that are part of standard Python3's random module
but are omitted from MicroPython's implementation

Specifically _not_ called "random.py" to prevent clobbering the standard random import
in other experimental libraries
"""

from math import log, sqrt
import random


def normal(mean=0.0, stdev=1.0):
    """
    Generate a random number with a normal distribution

    Uses the Marsaglia-Tsang algorithm, which isn't a perfectly normal
    distribution, but is likely close enough for EuroPi applications

    :param mean:   The desired mean for the distribution
    :param stdev:  The standard deviation of the distribution

    :return: A floating-point number chosen from a normal distribution with the
        given mean and standard deviation
    """
    x = 0.0
    y = 0.0
    z = 0.0
    while True:
        x = random.random() * 2.0 - 1.0
        y = random.random() * 2.0 - 1.0
        z = x**2 + y**2

        if 0 < z and z <= 1.0:
            break

    q = sqrt(-2.0 * log(z) / z)
    return mean + stdev * x * q


def shuffle(l):
    """
    Re-order the elements of a list in-situ

    :param l:  The list to shuffle
    """
    for i in range(len(l)):
        n = random.randint(i, len(l) - 1)
        tmp = l[i]
        l[i] = l[n]
        l[n] = tmp



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import europi
import gc
import machine
import sys
import time

from collections import OrderedDict
from europi import oled, OLED_HEIGHT, OLED_WIDTH, CHAR_HEIGHT, CHAR_WIDTH, reset_state
from europi_log import *
from europi_script import EuroPiScript
from ui import Menu

SCRIPT_DIR = "/lib/contrib/"
REG_FILE_CODE = 0x8000
DEBUG = False


class PrintMemoryUse:
    def __init__(self, label=""):
        self.label = label

    def __enter__(self):
        gc.collect()  # Note that preemptive GC is required to get all of the scripts loaded
        if DEBUG:
            self.before = gc.mem_free()

    def __exit__(self, *args):
        if DEBUG:
            gc.collect()
            after = gc.mem_free()
            log_info(
                f"free: {after/1024: >6.2f}k, used: {(self.before - after)/1024: >6.2f}k   {self.label}",
                "bootloader",
            )


class BootloaderMenu(EuroPiScript):
    """A Menu which allows the user to select and execute one of the scripts available to this EuroPi. The menu will
    include the ``EuroPiScripts`` provided to it in the constructor. Button press handlers are added that allow the user
    to exit the script and return to the menu by holding both buttons down for a short period of time. If the module is
    restarted while a script is running it will boot right back into that same script.

    In the menu:

    * **Button 1:** choose the selected item
    * **Button 2:** choose the selected item
    * **Knob 1:** unused
    * **Knob 2:** change the current selection

    In a program that was launched from the menu:

    * Hold both buttons for at least 0.5s and release to return to the menu.
    """

    def __init__(self, scripts):
        """Create the bootloader menu

        :param scripts:  Dictionary where the keys are the display names of the classes to include in the menu and the
            values are the fully-qualified names of the EuroPiScript classes that we launch
        """
        self.scripts = scripts
        self.run_request = None

    @staticmethod
    def show_progress(percentage):
        oled.hline(0, OLED_HEIGHT - 1, int(OLED_WIDTH * percentage), 1)
        oled.show()

    @staticmethod
    def get_class_for_name(script_class_name) -> type:
        try:
            module, clazz = script_class_name.rsplit(".", 1)
            return getattr(__import__(module, None, None, [None]), clazz)
        except Exception as e:
            log_warning(
                f"Warning: Ignoring bad qualified class name: {script_class_name}\n  caused by: {e}",
                "bootloader",
            )
            return None

    @classmethod
    def _is_europi_script(cls, c):
        return issubclass(c, EuroPiScript)

    @classmethod
    def show_error(cls, title, message, duration=1):
        """Show a brief error message on-screen saying we can't launch the requested script

        :param title:    The title of the error
        :param message:  The body of the error message
        :param duration: The number of seconds the message should show. If negative the message is shown forever
        """
        oled.centre_text(f"--{title}--\n{message}")

        if duration > 0:
            time.sleep(duration)

    def launch(self, selected_item):
        """Callback function for when the user chooses a menu item to launch

        Sets run_request to the name of the class to launch.  No validation is done here to keep this
        callback small
        """
        self.run_request = self.scripts[selected_item]

    def exit_to_menu(self):
        self.remove_state()
        # Attempt to save the state of this script if it has been implemented.
        self.save_state()  # TODO: isn't this the wrong state?
        machine.reset()  # why doesn't machine.soft_reset() work anymore?

    def run_menu(self) -> type:
        """Prompt the user to select a EuroPiScript class from the menu and return it

        If the class is not a valid EuroPiScript an error message is shown and selection continues

        :return: The type corresponding to self.run_request as set by the self.launch callback
        """
        self.menu = Menu(
            items=list(sorted(self.scripts.keys())),
            select_func=self.launch,
            select_knob=europi.k2,
            choice_buttons=[europi.b1, europi.b2],
        )

        # let the user make a selection
        # the menu selection returns the desired class name but doesn't validate it
        # that validation is handled here
        launch_class = None
        old_selected = -1
        while launch_class is None:
            if self.run_request:
                launch_class = self.get_class_for_name(self.run_request)
                if not self._is_europi_script(launch_class):
                    self.show_error("Launch Err", "Invalid script class")
                    launch_class = None
                    self.run_request = None
            else:
                if old_selected != self.menu.selected:
                    old_selected = self.menu.selected
                    self.menu.draw_menu()
                time.sleep(0.1)

        return launch_class

    def main(self):
        saved_state = self.load_state_json()
        script_class_name = saved_state.get("last_launched", None)
        script_class = None

        if script_class_name:
            script_class = self.get_class_for_name(script_class_name)

        if not script_class:
            script_class = self.run_menu()
            self.save_state_json(
                {"last_launched": f"{script_class.__module__}.{script_class.__name__}"}
            )
            machine.reset()
        else:
            # setup the exit handlers, and execute the selection
            europi.b1._handler_both(europi.b2, self.exit_to_menu)
            europi.b2._handler_both(europi.b1, self.exit_to_menu)

            try:
                if (
                    europi.europi_config.MENU_AFTER_POWER_ON
                    or script_class_name == "calibrate.Calibrate"
                ):
                    # Remove the last-launched file to force the module back to the menu after it powers-on next time
                    self.save_state_json({})

                script_class().main()
            except Exception as err:
                # set all outputs to zero for safety
                europi.turn_off_all_cvs()

                # in case we have the USB cable connected, print the stack trace for debugging
                # otherwise, just halt and show the error message
                log_error(f"Failed to run script: {err}", "bootloader")
                sys.print_exception(err)

                # show the type & first portion of the exception on the OLED
                # we can only fit so many characters, so truncate as needed
                MAX_CHARS = OLED_WIDTH // CHAR_WIDTH
                self.show_error(
                    "Crash", f"{err.__class__.__name__[0:MAX_CHARS]}\n{str(err)[0:MAX_CHARS]}", -1
                )

                # Log the crash to a file for later analysis/recovery
                try:
                    with open("last_crash.log", "w") as log_file:
                        log_file.write(f"{time.ticks_ms()}: {err}\n")
                        sys.print_exception(err, log_file)

                    log_error(f"Crash! See last_crash.txt for details: {err}", "bootloader")
                except:
                    # If we fail to create the error log, just silently fail; we don't need
                    # an additional exception to handle
                    pass



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from europi_hardware import Knob, MAX_UINT16

from experimental.math_extras import median


DEFAULT_THRESHOLD = 0.05


class LockableKnob(Knob):
    """A Knob whose state can be locked on the current value. Once locked, reading the knob's
    position will always return the locked value. When unlocking the knob, the value still doesn't
    change until the position of the knob moves to within ``threshold`` percent of the value. This
    prevents large jumps in a stetting when the knob is unlocked.

    This class is useful for cases where you want to have a single physical knob control several
    parameters (see also the :class:`KnobBank` class). Or where the value of a parameter needs to be
    disassociated from the position of the knob, as in after loading saved state.

    This class accepts two different parameters to specify it's initial value,
    `initial_uint16_value` and `initial_percentage_value`. Only one initial value may be specified.
    If both are specified, `initial_percentage_value` is ignored. The percentage value is more
    useful if you would like to hardcode a starting value for a knob in the code in a readable way.
    The uint16 value uses the knob's internal representation and is more appropriate for use when
    loading a saved knob position. If your script would like to read the internal representation of
    current position of a `LockableKnob`, first lock the knob, then read it's value.::

        lockable_knob.lock()
        internal_rep = lockable_knob.value

    :param knob: The knob to wrap.
    :param initial_uint16_value: The UINT16 (0-`europi_hardware.MAXINT16`) value to lock the knob at. If a value is provided the new knob is locked, otherwise it is unlocked.
    :param initial_percentage_value: The percentage (as a decimal 0-1) value to lock the knob at. If a value is provided the new knob is locked, otherwise it is unlocked.
    :param threshold: a decimal between 0 and 1 representing how close the knob must be to the locked value in order to unlock. The percentage is in terms of the knobs full range. Defaults to 5% (0.05)
    """

    STATE_UNLOCKED = 0
    STATE_UNLOCK_REQUESTED = 1
    STATE_LOCKED = 2

    def __init__(
        self,
        knob: Knob,
        initial_percentage_value=None,
        initial_uint16_value=None,
        threshold_percentage=DEFAULT_THRESHOLD,
    ):
        super().__init__(knob.pin_id)
        self.pin = knob.pin  # Share the ADC

        if initial_uint16_value != None:
            self.value = initial_uint16_value
            self.state = LockableKnob.STATE_LOCKED
        elif initial_percentage_value != None:
            self.value = (1 - initial_percentage_value) * MAX_UINT16
            self.state = LockableKnob.STATE_LOCKED
        else:
            self.value = MAX_UINT16  # Min value
            self.state = LockableKnob.STATE_UNLOCKED

        self.threshold = int(threshold_percentage * MAX_UINT16)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.pin}, {self.value}, {self.state})"

    def _sample_adc(self, samples=None):
        if self.state == LockableKnob.STATE_LOCKED:
            return self.value

        elif self.state == LockableKnob.STATE_UNLOCKED:
            return super()._sample_adc(samples)

        else:  # STATE_UNLOCK_REQUESTED:
            current_value = super()._sample_adc(samples)
            if abs(self.value - current_value) < self.threshold:
                self.state = LockableKnob.STATE_UNLOCKED
                return current_value
            else:
                return self.value

    def lock(self, samples=None):
        """Locks this knob at its current state. Makes a call to the underlying knob's
        ``_sample_adc()`` method with the given number of samples."""
        self.value = self._sample_adc(samples=samples)
        self.state = LockableKnob.STATE_LOCKED

    def change_lock_value(self, percent=0.0):
        """Change the current value to reflect a desired percentage

        :param percent:  The 0-1 value we want to re-lock the knob at
        """
        self.value = int(MAX_UINT16 * (1.0 - percent))

    def request_unlock(self):
        """Requests that the knob be unlocked. The knob will unlock the next time a reading of it's
        position is taken that is withing the threshold percentage of the locked value. That is,
        when the knob is moved close to the locked value. If :meth:`~LockableKnob.lock()` is called
        before the knob unlocks, the unlock is aborted.
        """
        if self.state == LockableKnob.STATE_LOCKED:
            self.state = LockableKnob.STATE_UNLOCK_REQUESTED


class DisabledKnob(LockableKnob):
    """A :class:`LockableKnob` that cannot be unlocked and whose value is unimportant. Useful when
    building multifunction knobs.

    :param knob: The knob to wrap."""

    def __init__(self, knob: Knob):
        super().__init__(knob, initial_uint16_value=MAX_UINT16)

    def request_unlock(self):
        """LockedKnob can never be unlocked"""
        return


class KnobBank:
    """A ``KnobBank`` is a group of named 'virtual' :class:``LockableKnobs`` that share the same
    physical knob. Only one of these knobs is active (unlocked) at a time. This allows for a single
    knob to be used to control many parameters.

    It is recommended that ``KnobBanks`` be created using the :class:``Builder``, as in::

       k1_bank = (
           KnobBank.builder(k1)
           .with_disabled_knob()
           .with_unlocked_knob("x", threshold=0.02)
           .with_locked_knob("y", initial_percentage_value=1)
           .build()
       )

    Knobs can be referenced using their names::

       k1_bank.x.percent()
       k1_bank.y.read_position()

    .. note::
       ``KnobBank`` is not thread safe. It is possible to end up with two unlocked knobs if you call
       :meth:`~KnobBank.next()` and take readings from a knob in two different threads. This can
       easily happen if your script calls ``next()`` in a button handler, while constantly reading
       the knob state in a main loop. The workaround is to set a flag in the button handler and call
       ``next()`` in the main loop.

       For example::

          class KnobBankExample(EuroPiScript):
              def __init__(self):
                  super().__init__()
                  self.next_k1 = False

                  self.kb1 = (
                      KnobBank.builder(k1)
                      .with_locked_knob("p1", initial_percentage_value=1, threshold_percentage=0.02)
                      .with_locked_knob("p2", initial_percentage_value=1)
                      .with_locked_knob("p3", initial_percentage_value=1)
                      .build()
                  )

                  @b1.handler
                  def next_knob1():
                      self.next_k1 = True

              def main(self):

                  while True:
                      if self.next_k1:
                          self.kb1.next()
                          self.next_k1 = False

                      # main loop body
    """

    def __init__(self, physical_knob: Knob, virtual_knobs, initial_selection) -> None:
        self.index = 0
        self.knobs = []
        self.names = []

        for name, knob in virtual_knobs.items():
            setattr(self, name, knob)  # make knob available by its name
            self.knobs.append(knob)
            self.names.append(name)
        self.knobs[initial_selection].request_unlock()
        self.index = initial_selection

    @property
    def current(self) -> LockableKnob:
        """The currently active knob."""
        return self.knobs[self.index]

    @property
    def current_name(self) -> str:
        return self.names[self.index]

    def next(self):  # potential race condition: next() and _sample_adc() can both change state
        """Select the next knob by locking the current knob, and requesting an unlock on the next in
        the bank."""
        self.current.lock()
        self.index = (self.index + 1) % len(self.knobs)
        self.current.request_unlock()

    def set_current(self, name):
        """Set the currently-unlocked knob to the one whose name matches the argument

        :param name:  The name of the knob to set as the current one
        """
        try:
            index = self.names.index(name)
            self.current.lock()
            self.index = index
            self.current.request_unlock()
        except ValueError:
            # if the name isn't found, just silently trap the exception
            pass

    def __getitem__(self, name):
        """Get the LockableKnob in this bank with the given name

        :param name:  The name of the knob to return, or None if the name isn't found
        """
        try:
            index = self.names.index(name)
            return self.knobs[index]
        except ValueError:
            return None

    class Builder:
        """A convenient interface for creating a :class:`KnobBank` with consistent initial state."""

        def __init__(self, knob: Knob) -> None:
            self.knob = knob
            self.knobs_by_name = OrderedDict()
            self.initial_index = None

        def with_disabled_knob(self) -> "Builder":
            """Add a :class:`DisabledKnob` to the bank. This disables the knob so that no parameters can
            be changed."""
            self.knobs_by_name[DisabledKnob.__name__] = DisabledKnob(self.knob)
            return self

        def with_locked_knob(
            self,
            name: str,
            initial_percentage_value=None,
            initial_uint16_value=None,
            threshold_percentage=None,
            threshold_from_choice_count=None,
        ) -> "Builder":
            """Add a :class:`LockableKnob` to the bank whose initial state is locked.

            `threshold_from_choice_count` is a convenience parameter to be used in the case where
            this knob will be used to select from a relatively few number of choices, via the
            :meth:`~europi_hardware.Knob.choice()` method. Pass the number of choices to this parameter and
            an appropriate threshold value will be calculated.

            :param name: the name of this virtual knob
            :param threshold_percentage: the threshold percentage for this knob as described by :class:`LockableKnob`
            :param threshold_from_choice_count: Provides the number of choices this knob will be used with in order to generate an appropriate threshold.
            """
            if initial_uint16_value is None and initial_percentage_value is None:
                raise ValueError(
                    "initial_percentage_value and initial_uint16_value cannot both be None"
                )

            return self._with_knob(
                name,
                initial_percentage_value=initial_percentage_value,
                initial_uint16_value=initial_uint16_value,
                threshold_percentage=threshold_percentage,
                threshold_from_choice_count=threshold_from_choice_count,
            )

        def with_unlocked_knob(
            self,
            name: str,
            threshold_percentage=None,
            threshold_from_choice_count=None,
        ) -> "Builder":
            """Add a :class:`LockableKnob` to the bank whose initial state is unlocked. This knob
            will be active. Only one unlocked knob may be added to the bank.

            `threshold_from_choice_count` is a convenience parameter to be used in the case where
            this knob will be used to select from a relatively few number of choices, via the
            :meth:`~europi_hardware.Knob.choice()` method. Pass the number of choices to this parameter and
            an appropriate threshold value will be calculated.

            :param name: the name of this virtual knob
            :param threshold_percentage: the threshold percentage for this knob as described by :class:`LockableKnob`
            :param threshold_from_choice_count: Provides the number of choices this knob will be used with in order to generate an appropriate threshold.

            """
            if self.initial_index != None:
                raise ValueError(f"Second unlocked knob specified: {name}")

            self._with_knob(name, None, None, threshold_percentage, threshold_from_choice_count)

            self.initial_index = len(self.knobs_by_name) - 1

            return self

        def _with_knob(
            self,
            name: str,
            initial_percentage_value,
            initial_uint16_value,
            threshold_percentage,
            threshold_from_choice_count=None,
        ):
            if name == None:
                raise ValueError("Knob name cannot be None")

            if threshold_percentage != None and threshold_from_choice_count != None:
                raise ValueError(
                    "Specify only one of either threshold_percentage or threshold_from_choice_count not both"
                )

            if threshold_from_choice_count != None:
                threshold_percentage = 1 / threshold_from_choice_count

            elif threshold_percentage == None:
                threshold_percentage = DEFAULT_THRESHOLD

            self.knobs_by_name[name] = LockableKnob(
                self.knob,
                initial_percentage_value=initial_percentage_value,
                initial_uint16_value=initial_uint16_value,
                threshold_percentage=threshold_percentage,
            )

            return self

        def build(self) -> "KnobBank":
            """Create the :class:`KnobBank` with the specified knobs."""
            if len(self.knobs_by_name) < 0:
                raise ValueError(f"Must specify at least one knob in the bank.")
            if self.initial_index == None:
                self.initial_index = 0
            return KnobBank(
                physical_knob=self.knob,
                virtual_knobs=self.knobs_by_name,
                initial_selection=self.initial_index,
            )

    @staticmethod
    def builder(knob: Knob) -> Builder:
        return KnobBank.Builder(knob)


class BufferedKnob(Knob):
    """A wrapper for a Knob instance whose value remains fixed until .update(...) is called

    This allows multiple uses of .percent(), .choice(...), etc... without forcing a re-read of
    the ADC value

    :param knob: The analogue input to wrap e.g. ``europi_hardware.k1``
    """

    def __init__(self, knob):
        super().__init__(knob.pin_id)
        self.value = 0

    def _sample_adc(self, samples=None):
        """Overrides the internal function that samples the ADC

        Instead of sampling we simply return the most recent sample value

        :param samples:  Ignored, but needed by the _sample_adc API used by AnalogueReader
        """
        return self.value

    def update(self, samples=None):
        """Re-read the ADC and update the buffered value

        :param samples:  Specifies the number of samples to average to de-noise the ADC reading
            See europi_hardware.AnalogueReader for details on ADC sampling
        """
        self.value = super()._sample_adc(samples)


class MedianAnalogInput:
    """A wrapper for an analogue input (e.g. knob, ain) that provides additional smoothing & debouncing

    This class uses a window of the n latest samples from the underlying input and uses the median of
    those samples. Larger window sizes will produce a more stable output, but at the cost of slowing
    down reaction time to changes.

    :param analog_in:    The input we're wrapping (e.g. k1, k2, ain)
    :param samples:      The number of samples to use when reading from analog_in
    :param window_size:  The number of samples used for calculating the median
    """

    def __init__(self, analog_in, samples=100, window_size=5):
        self.analog_in = analog_in
        self.n_samples = samples
        self.window_size = window_size

        self.samples = []

    def percent(self):
        """Read the hardware percentage and apply our additional smoothing

        Smoothing is done using a simple 5-window median filter
        """
        self.samples.append(self.analog_in.percent(self.n_samples))

        if len(self.samples) > self.window_size:
            self.samples.pop(0)

        return median(self.samples)



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript

from configuration import ConfigFile

from tools.conf_edit import k2_bank, SectionHeader
from experimental.experimental_config import *
from experimental.knobs import KnobBank
from experimental.settings_menu import *

from framebuf import FrameBuffer, MONO_HLSB


class ExperimentalConfigurationEditor(EuroPiScript):
    """
    A global configuration-editor menu

    Used to edit the global EuroPiConfig settings -- use at your own peril!
    """

    def __init__(self):
        super().__init__()

        # fmt: off
        boolean_labels = {
            True: "Yes",
            False: "No"
        }

        quantizer_items = []
        rtc_items = []
        wifi_items = []

        config_points = ExperimentalConfig.config_points()
        for cfg in config_points:
            if type(cfg) is BooleanConfigPoint:
                labels = boolean_labels
            else:
                labels = None

            if cfg.name == "VOLTS_PER_OCTAVE":
                title = "V/Oct"
                prefix = "Quant"
                items = quantizer_items
            elif cfg.name == "ENABLE_WEBREPL":
                title = "WebREPL"
                prefix = "WiFi"
                items = wifi_items
            elif "RTC" in cfg.name or "UTC" in cfg.name:
                title = cfg.name.replace("RTC", "").replace("_", ' ').lower().strip()
                prefix = "RTC"
                items = rtc_items
            elif "WIFI" in cfg.name:
                title = cfg.name.replace("WIFI", "").replace("_", ' ').lower().strip()
                prefix = "WiFi"
                items = wifi_items

            items.append(
                SettingMenuItem(
                    config_point=cfg,
                    labels=labels,
                    float_resolution=1,
                    prefix=prefix,
                    title=title,
                )
            )

        self.menu = SettingsMenu(
            menu_items=[
                # Voltage properties
                SectionHeader(
                    title="Quantizer",
                    children=quantizer_items,
                ),

                # System properties
                SectionHeader(
                    title="Realtime Clk",
                    children=rtc_items,
                ),

                # Display properties
                SectionHeader(
                    title="WiFi",
                    children=wifi_items,
                ),
            ],
            navigation_knob=k2_bank,
        )
        self.menu.load_defaults(ConfigFile.config_filename(ExperimentalConfig))
        # fmt: on

    def main(self):
        # fmt: off
        disk_icon = bytearray(b'\xff\xf0\x9f\xe8\x9f$\x9f"\x9f!\x9f\xe1\x80\x01\xbf\xfd\xa0\x05\xaf\xf5\xa0\x05\xaf\xf5\xa0\x05\xbf\xfd\x80\x01\xff\xff')
        # fmt: on

        icon_height = OLED_HEIGHT // 2 - 8
        text_height = OLED_HEIGHT // 2 - CHAR_HEIGHT // 2

        while True:
            if self.menu.ui_dirty:
                oled.fill(0)
                self.menu.draw()
                oled.show()

            if self.menu.settings_dirty:
                # visually indicate we're saving
                oled.fill(0)
                oled.blit(FrameBuffer(disk_icon, 16, 16, MONO_HLSB), 0, icon_height)
                oled.text("Saving...", 18, text_height, 1)
                oled.show()

                self.menu.save(ConfigFile.config_filename(ExperimentalConfig))
                time.sleep(0.5)

                oled.fill(0)
                self.menu.draw()
                oled.show()


if __name__ == "__main__":
    ExperimentalConfigurationEditor().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for converting analogue signals to digital inputs"""
from europi import HIGH, LOW

import utime


class AnalogReaderDigitalWrapper:
    """Wraps an AnalogReader to allow it to simulate a DigitalReader.

    The EuroPiScript class using the AnalogReaderDigitalWrapper must call `.update()` to read the current state of
    the analogue input and trigger any rising/falling edge callbacks.

    The value returned by `.value()` is accurate to the last time `.update()` was called.

    :param ain:  The AnalogReader we're wrapping
    :param debounce:  The number of consecutive high/low signals needed to flip the digital state
    :param high_low_cutoff:  The threshold at which the analog signal is considered high
    :param cb_rising:  A function to call on the rising edge of the signal
    :param cb_falling:  A function to call on the falling edge of the signal
    """

    def __init__(
        self, ain, debounce=1, high_low_cutoff=0.8, cb_rising=lambda: None, cb_falling=lambda: None
    ):
        self.ain = ain
        self.debounce = debounce
        self.high_low_cutoff = high_low_cutoff
        self.last_rising_time = 0
        self.last_falling_time = 0
        self.debounce_counter = 0
        self.state = False

        if not callable(cb_rising) or not callable(cb_falling):
            raise ValueError("Provided callback func is not callable")

        self.cb_rising = cb_rising
        self.cb_falling = cb_falling

    def value(self):
        """Returns europi.HIGH or europi.LOW depending on the state of the input"""
        return HIGH if self.state else LOW

    def update(self):
        """Reads the current value of the analogue input and updates the internal state"""
        volts = self.ain.read_voltage()

        # count how many opposite-voltage readings we have
        if (self.state and volts < self.high_low_cutoff) or (
            not self.state and volts >= self.high_low_cutoff
        ):
            self.debounce_counter += 1

        # change state if we've reached the debounce threshold
        if self.debounce_counter >= self.debounce:
            self.debounce_counter = 0
            self.state = not self.state
            if self.state:
                self.last_rising_time = utime.ticks_ms()
                self.cb_rising()
            else:
                self.last_falling_time = utime.ticks_ms()
                self.cb_falling()

    def last_rising_ms(self):
        return self.last_rising_time

    def last_falling_ms(self):
        return self.last_falling_time

    def handler(self, func):
        """Define the callback function to call when rising edge detected."""
        if not callable(func):
            raise ValueError("Provided handler func is not callable")
        self.cb_rising = func

    def handler_falling(self, func):
        """Define the callback function to call when falling edge detected."""
        if not callable(func):
            raise ValueError("Provided handler func is not callable")
        self.cb_falling = func



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface class for the DS3231 Realtime Clock

This class is designed to work with a DS3231 chip mounted on an I2C carrier board
that can be connected to EuroPi's external I2C interface. The user is required to
1) provide their own RTC module
2) create/source an appropriate adapter to connect the GND, VCC, SDA, and SCL pins on EuroPi to the RTC module
3) Mount the RTC module securely in such a way that it won't come loose nor accidentally short out any other components.

Compatible RTC modules can be purchased relatively cheaply online. e.g.:
- https://www.amazon.ca/DS3231-Precision-AT24C32-Arduino-Raspberry/dp/B07V68443F (not an afficliate link)

Based on code by Willem Peterse released under the MIT license (c) 2020:
https://github.com/pangopi/micropython-DS3231-AT24C32/blob/main/ds3231.py

in-turn based on work by Mike Causer for the DS1307:
https://github.com/mcauser/micropython-tinyrtc-i2c/blob/master/ds1307.py

If you have a brand-new DS3231, or have recently replaced your module's battery, you may need to manually update
the clock.  To do so:

#. Connect the clock module to your EuroPi
#. Connect your EuroPi to a computer via the USB port
#. Open Thonny and make sure experimental_config is configured to use the DS3231. If you make any changes to experimental_config, restart the Raspberry Pi Pico before proceeding.
#. In Thonny's Python terminal, run the following code:

.. code-block::

    >>> from experimental.rtc import clock
    >>> clock.source.set_datetime((2025, 6, 14, 22, 59, 0, 6, 0))


This will set the clock to 14 June 2025, 22:59:00, and set the weekday to Saturday (6).
The tuple is of the form (Year, Month, Day, Hour, Minute, Second, Weekday, Yearday). It is recommended
to set the seconds & weekday, but it is optional.

Note that the clock _should_ be set to UTC, not local time. If you choose to use local time instead
some scripts that assume the clock is set to UTC may behave incorrectly.
"""

from europi_log import *

from experimental.clocks.clock_source import ExternalClockSource

try:
    from micropython import const
except ImportError:

    def const(x):
        return x


# fmt: off
DATETIME_REG    = const(0)   # 7 bytes
ALARM1_REG      = const(7)   # 5 bytes
ALARM2_REG      = const(11)  # 4 bytes
CONTROL_REG     = const(14)
STATUS_REG      = const(15)
AGING_REG       = const(16)
TEMPERATURE_REG = const(17)  # 2 bytes
# fmt: on


def dectobcd(decimal):
    """Convert decimal to binary coded decimal (BCD) format"""
    return (decimal // 10) << 4 | (decimal % 10)


def bcdtodec(bcd):
    """Convert binary coded decimal to decimal"""
    return ((bcd >> 4) * 10) + (bcd & 0x0F)


class DS3231(ExternalClockSource):
    """
    DS3231 RTC driver.

    Hard coded to work with year 2000-2099.
    """

    # fmt: off
    FREQ_1      = const(1)
    FREQ_1024   = const(2)
    FREQ_4096   = const(3)
    FREQ_8192   = const(4)
    SQW_32K     = const(1)

    AL1_EVERY_S     = const(15)  # Alarm every second
    AL1_MATCH_S     = const(14)  # Alarm when seconds match (every minute)
    AL1_MATCH_MS    = const(12)  # Alarm when minutes, seconds match (every hour)
    AL1_MATCH_HMS   = const(8)   # Alarm when hours, minutes, seconds match (every day)
    AL1_MATCH_DHMS  = const(0)   # Alarm when day|wday, hour, min, sec match (specific wday / mday) (once per month/week)

    AL2_EVERY_M     = const(7)   # Alarm every minute on 00 seconds
    AL2_MATCH_M     = const(6)   # Alarm when minutes match (every hour)
    AL2_MATCH_HM    = const(4)   # Alarm when hours and minutes match (every day)
    AL2_MATCH_DHM   = const(0)   # Alarm when day|wday match (once per month/week)
    # fmt: on

    def __init__(self, i2c, addr=0x68):
        # fmt: off
        super().__init__()
        self.i2c = i2c
        self.addr = addr
        self._timebuf = bytearray(7)  # Pre-allocate a buffer for the time data
        self._buf = bytearray(1)      # Pre-allocate a single bytearray for re-use
        self._al1_buf = bytearray(4)
        self._al2buf = bytearray(3)
        # fmt: on

    def datetime(self):
        """
        Get the current time.

        Returns in 24h format, converts to 24h if clock is set to 12h format

        :return: a tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        self.i2c.readfrom_mem_into(self.addr, DATETIME_REG, self._timebuf)
        # 0x00 - Seconds    BCD
        # 0x01 - Minutes    BCD
        # 0x02 - Hour       0 12/24 AM/PM/20s BCD
        # 0x03 - WDay 1-7   0000 0 BCD
        # 0x04 - Day 1-31   00 BCD
        # 0x05 - Month 1-12 Century 00 BCD
        # 0x06 - Year 0-99  BCD (2000-2099)
        seconds = bcdtodec(self._timebuf[0])
        minutes = bcdtodec(self._timebuf[1])

        # Check for 12 hour mode bit
        if (self._timebuf[2] & 0x40) >> 6:
            hour = bcdtodec(self._timebuf[2] & 0x9F)  # Mask out bit 6(12/24) and 5(AM/PM)
            if (self._timebuf[2] & 0x20) >> 5:  # bit 5(AM/PM)
                # PM
                hour += 12
        else:
            # 24h mode
            hour = bcdtodec(self._timebuf[2] & 0xBF)  # Mask bit 6 (12/24 format)

        weekday = bcdtodec(self._timebuf[3])  # Can be set arbitrarily by user (1,7)
        day = bcdtodec(self._timebuf[4])
        month = bcdtodec(self._timebuf[5] & 0x7F)  # Mask out the century bit
        year = bcdtodec(self._timebuf[6]) + 2000

        if self.OSF():
            log_warning("Oscillator stop flag set. Time may not be accurate", "ds3231")

        return (
            year,
            month,
            day,
            hour,
            minutes,
            seconds,
            weekday,
            0,
        )

    def set_datetime(self, datetime):
        """
        Set the current time.

        :param datetime: tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        self.check_valid_datetime(datetime)

        # fmt: off
        try:
            self._timebuf[3] = dectobcd(datetime[6])             # Day of week
        except IndexError:
            self._timebuf[3] = 0
        try:
            self._timebuf[0] = dectobcd(datetime[5])             # Seconds
        except IndexError:
            self._timebuf[0] = 0
        self._timebuf[1] = dectobcd(datetime[4])                 # Minutes
        self._timebuf[2] = dectobcd(datetime[3])                 # Hour + the 24h format flag
        self._timebuf[4] = dectobcd(datetime[2])                 # Day
        self._timebuf[5] = dectobcd(datetime[1]) & 0xFF          # Month + mask the century flag
        self._timebuf[6] = dectobcd(int(str(datetime[0])[-2:]))  # Year can be yyyy, or yy
        self.i2c.writeto_mem(self.addr, DATETIME_REG, self._timebuf)
        self._OSF_reset()
        return True
        # fmt: on

    def square_wave(self, freq=None):
        """
        Outputs Square Wave Signal

        The alarm interrupts are disabled when enabling a square wave output. Disabling SWQ out does
        not enable the alarm interrupts. Set them manually with the alarm_int() method.

        :param freq: specify the frequency:
            * None: returns current setting
            * False = disable SQW output,
            * 1 =     1 Hz,
            * 2 = 1.024 kHz,
            * 3 = 4.096 kHz,
            * 4 = 8.192 kHz
        """
        # fmt: off
        if freq is None:
            return self.i2c.readfrom_mem(self.addr, CONTROL_REG, 1)[0]

        if not freq:
            # Set INTCN (bit 2) to 1 and both ALIE (bits 1 & 0) to 0
            self.i2c.readfrom_mem_into(self.addr, CONTROL_REG, self._buf)
            self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([(self._buf[0] & 0xF8) | 0x04]))
        else:
            # Set the frequency in the control reg and at the same time set the INTCN to 0
            freq -= 1
            self.i2c.readfrom_mem_into(self.addr, CONTROL_REG, self._buf)
            self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([(self._buf[0] & 0xE3) | (freq << 3)]))
        return True
        # fmt: on

    def alarm1(self, time=None, match=AL1_MATCH_DHMS, int_en=True, weekday=False):
        """Set alarm1, can match mday, wday, hour, minute, second

        :param time: tuple, (second,[ minute[, hour[, day]]])
        :param weekday: bool, select mday (False) or wday (True)
        :param match: int, match const
        :param int_en: bool, enable interrupt on alarm match on SQW/INT pin (disables SQW output)"""
        if time is None:
            # TODO Return readable string
            self.i2c.readfrom_mem_into(self.addr, ALARM1_REG, self._al1_buf)
            return self._al1_buf

        if isinstance(time, int):
            time = (time,)

        a1m4 = (match & 0x08) << 4
        a1m3 = (match & 0x04) << 5
        a1m2 = (match & 0x02) << 6
        a1m1 = (match & 0x01) << 7

        dydt = (1 << 6) if weekday else 0  # day / date bit

        # fmt: off
        self._al1_buf[0] = dectobcd(time[0]) | a1m1                                             # second
        self._al1_buf[1] = (dectobcd(time[1]) | a1m2) if len(time) > 1 else a1m2                # minute
        self._al1_buf[2] = (dectobcd(time[2]) | a1m3) if len(time) > 2 else a1m3                # hour
        self._al1_buf[3] = (dectobcd(time[3]) | a1m4 | dydt) if len(time) > 3 else a1m4 | dydt  # day (wday|mday)
        # fmt: on

        self.i2c.writeto_mem(self.addr, ALARM1_REG, self._al1_buf)

        # Set the interrupt bit
        self.alarm_int(enable=int_en, alarm=1)

        # Check the alarm (will reset the alarm flag)
        self.check_alarm(1)

        return self._al1_buf

    def alarm2(self, time=None, match=AL2_MATCH_DHM, int_en=True, weekday=False):
        """Get/set alarm 2 (can match minute, hour, day)

        :param time: tuple, (minute[, hour[, day]])
        :param weekday: bool, select mday (False) or wday (True)
        :param match: int, match const
        :param int_en: bool, enable interrupt on alarm match on SQW/INT pin (disables SQW output)
        :return: bytearray(3), the alarm settings register"""
        if time is None:
            # TODO Return readable string
            self.i2c.readfrom_mem_into(self.addr, ALARM2_REG, self._al2buf)
            return self._al2buf

        if isinstance(time, int):
            time = (time,)

        a2m4 = (match & 0x04) << 5  # masks
        a2m3 = (match & 0x02) << 6
        a2m2 = (match & 0x01) << 7

        # fmt: off
        dydt = (1 << 6) if weekday else 0  # day / date bit
        self._al2buf[0] = dectobcd(time[0]) | a2m2 if len(time) > 1 else a2m2                # minute
        self._al2buf[1] = dectobcd(time[1]) | a2m3 if len(time) > 2 else a2m3                # hour
        self._al2buf[2] = dectobcd(time[2]) | a2m4 | dydt if len(time) > 3 else a2m4 | dydt  # day
        # fmt: on

        self.i2c.writeto_mem(self.addr, ALARM2_REG, self._al2buf)

        # Set the interrupt bits
        self.alarm_int(enable=int_en, alarm=2)

        # Check the alarm (will reset the alarm flag)
        self.check_alarm(2)

        return self._al2buf

    def alarm_int(self, enable=True, alarm=0):
        """Enable/disable interrupt for alarm1, alarm2 or both.

        Enabling the interrupts disables the SQW output

        :param enable: bool, enable/disable interrupts
        :param alarm: int, alarm nr (0 to set both interrupts)
        :return: the control register"""
        # fmt: off
        if alarm in (0, 1):
            self.i2c.readfrom_mem_into(self.addr, CONTROL_REG, self._buf)
            if enable:
                self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([(self._buf[0] & 0xFA) | 0x05]))
            else:
                self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([self._buf[0] & 0xFE]))

        if alarm in (0, 2):
            self.i2c.readfrom_mem_into(self.addr, CONTROL_REG, self._buf)
            if enable:
                self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([(self._buf[0] & 0xF9) | 0x06]))
            else:
                self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([self._buf[0] & 0xFD]))

        return self.i2c.readfrom_mem(self.addr, CONTROL_REG, 1)
        # fmt: on

    def check_alarm(self, alarm):
        """Check if the alarm flag is set and clear the alarm flag"""
        self.i2c.readfrom_mem_into(self.addr, STATUS_REG, self._buf)
        if (self._buf[0] & alarm) == 0:
            # Alarm flag not set
            return False

        # Clear alarm flag bit
        self.i2c.writeto_mem(self.addr, STATUS_REG, bytearray([self._buf[0] & ~alarm]))
        return True

    def output_32kHz(self, enable=True):
        """Enable or disable the 32.768 kHz square wave output"""
        status = self.i2c.readfrom_mem(self.addr, STATUS_REG, 1)[0]
        if enable:
            self.i2c.writeto_mem(self.addr, STATUS_REG, bytearray([status | (1 << 3)]))
        else:
            self.i2c.writeto_mem(self.addr, STATUS_REG, bytearray([status & (~(1 << 3))]))

    def OSF(self):
        """Returns the oscillator stop flag (OSF).

        1 indicates that the oscillator is stopped or was stopped for some
        period in the past and may be used to judge the validity of
        the time data.
        returns : bool"""
        return bool(self.i2c.readfrom_mem(self.addr, STATUS_REG, 1)[0] >> 7)

    def _OSF_reset(self):
        """Clear the oscillator stop flag (OSF)"""
        self.i2c.readfrom_mem_into(self.addr, STATUS_REG, self._buf)
        self.i2c.writeto_mem(self.addr, STATUS_REG, bytearray([self._buf[0] & 0x7F]))

    def _is_busy(self):
        """Returns True when device is busy doing TCXO management"""
        return bool(self.i2c.readfrom_mem(self.addr, STATUS_REG, 1)[0] & (1 << 2))



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup

__version__ = ""
exec(open("./version.py").read())


def build_firmware_module_list():
    import os

    dirs = {"experimental", "tools"}
    excluded = {"__init__.py", "setup.py"}

    # grab files from the local directory first
    modules = [f[:-3] for f in os.listdir(".") if f.endswith(".py") and not f in excluded]

    # add any modules in additional directories
    for d in dirs:
        for root_dir, subdirs, files in os.walk(d):
            for f in files:
                if f.endswith(".py") and not f in excluded:
                    modules.append(root_dir.replace("/", ".") + "." + f[:-3])

    return modules


setup(
    name="micropython-europi",
    version=__version__,
    description="EuroPi module for MicroPython",
    long_description="""The EuroPi is a fully user reprogrammable EuroRack module based on the Raspberry Pi Pico, which allows users to process inputs and controls to produce outputs based on code written in Python.""",
    url="https://github.com/Allen-Synthesis/EuroPi",
    author="Allen Synthesis",
    author_email="contact@allensynthesis.co.uk",
    license="Apache 2.0",
    py_modules=build_firmware_module_list(),
    install_requires=[],
)



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import machine
from europi_script import EuroPiScript
from europi import b1, b2, oled
from europi_log import log_warning
from os import remove


class BootloaderMode(EuroPiScript):
    @classmethod
    def display_name(cls):
        """Push this script to the end of the menu."""
        return "~Bootloader Mode"

    def back(self):
        try:
            remove(
                "saved_state_BootloaderMenu.txt"
            )  # This file needs to be removed so that the menu is returned to rather than this script once it's entered bootloader mode
        except OSError:
            log_warning("OSError: File Not Found", "bootloader")
        machine.reset()  # Restart the module so that it will boot back into the menu

    def enter_bootloader(self):
        machine.bootloader()  # Enter bootloader mode

    def main(self):
        b1.handler(self.back)  # Button 1 returns to the menu system
        b2.handler(self.enter_bootloader)  # Button 2 sends the module into bootloader mode

        oled.fill(0)
        oled.text("B1 = Back", 0, 6)
        oled.text("B2 = Bootloader", 0, 18)
        oled.show()


if __name__ == "__main__":
    BootloaderMode().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Before using any of this library, follow the instructions in
`programming_instructions.md <https://github.com/Allen-Synthesis/EuroPi/blob/main/software/programming_instructions.md>`_
to set up your module.

The EuroPi library is a single file named europi.py. It should be imported into any custom program by using ``from europi import *`` to give you full access to the functions within, which are outlined below. Inputs and outputs are used as objects, which each have methods to allow them to be used. These methods are used by using the name of the object, for example 'cv3' followed by a '.' and then the method name, and finally a pair of brackets containing any parameters that the method requires.

For example::

    cv3.voltage(4.5)

Will set the CV output 3 to a voltage of 4.5V.
"""


import sys

from version import __version__

from configuration import ConfigSettings
from framebuf import FrameBuffer, MONO_HLSB

from europi_config import load_europi_config, MODEL_PICO_2W, MODEL_PICO_W
from europi_display import Display, DummyDisplay
from europi_hardware import *
from europi_log import *

from experimental.experimental_config import load_experimental_config
from experimental.wifi import WifiConnection, WifiError


if sys.implementation.name == "micropython":
    TEST_ENV = False  # We're in micropython, so we can assume access to real hardware
else:
    TEST_ENV = True  # This var is set when we don't have any real hardware, for example in a test or doc generation setting

# Initialize EuroPi global singleton instance variables
europi_config = load_europi_config()
experimental_config = load_experimental_config()

# OLED component display dimensions.
OLED_WIDTH = europi_config.DISPLAY_WIDTH
OLED_HEIGHT = europi_config.DISPLAY_HEIGHT

# Default font is 8x8 pixel monospaced font.
CHAR_WIDTH = 8
CHAR_HEIGHT = 8


# Helper functions.


def turn_off_all_cvs():
    """Calls cv.off() for every cv in cvs. This is done commonly enough in
    contrib scripts that a function seems useful.
    """
    for cv in cvs:
        cv.off()


def reset_state():
    """Return device to initial state with all components off and handlers reset."""
    if not TEST_ENV:
        oled.fill(0)
        oled.show()
    turn_off_all_cvs()
    for d in (b1, b2, din):
        d.reset_handler()


def bootsplash():
    """Display the EuroPi version when booting."""
    image = b"\x00\x00\x00\x03\xe0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\xf8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x1c\x00\x00\x00\x00\x00\x00\x00\x01\x80\x00\x00\x00\x00\x00\x07\x0f\xe0\x00\x00\x00\x00\x00\x00\x01\x80\x00\x00\x00\x00\x00\x03\x07\xf0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x800\x00\xc0\x00\x00\x0e\x01\xc0\x00\x00\x00\x00\x00\x07\xe0\xc0p\x03\xf0\xc1\x971\x8e1\x80\x00\x00\x00\x00\x1f\xf8\xc1\xe0\x06\x18\xc1\x98`\xcc\x19\x80\x00\x00\x00\x00\x1c\x1f\xc3\x83\x04\x0c\xc1\x98@L\t\x80\x00\x00\x00\x000\x07\x87\x07\x0c\x0c\xc1\x90@H\t\x80\x00\x00\x00\x000\x00\x0e\x0e\x0f\xf8\xc1\x90@H\t\x80\x00\x00\x00\x008\x00\x1c\x0c\x0c\x00\xc1\x90@H\t\x80\x00\x00\x00\x00\x1f\x808\x0e\x04\x00\xc1\x90`\xcc\x19\x80\x00\x00\x00\x00\x0f\xe01\xc7\x06\x18c\x101\x8e1\x80\x00\x00\x00\x00\x00ps\xe3\x01\xe0\x1c\x10\x0e\r\xc1\x80\x00\x00\x00\x00\x00\x18c1\x80\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x18\xe31\x80\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x008\xc1\x99\x80\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x0e1\xc1\x99\x80\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x1f\xe3\x80\xcf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00;\xc7\x00\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x000\x0e\x00@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x000\x1c\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00xp`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0\xf80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xc1\xd80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x83\x980\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x07\x180\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x86\x0c`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x0e\xe0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00|\x07\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    TH = bytearray(image)
    fb = FrameBuffer(TH, 128, 32, MONO_HLSB)
    oled.blit(fb, 0, 0)

    version_str = str(__version__)
    version_length = len(version_str)
    offset = int(((150 - (version_length * CHAR_WIDTH)) / 2))
    oled.text(version_str, offset, 20, 1)

    oled.show()


# Initialize the OLED
if not TEST_ENV:
    try:
        oled = Display(
            width=europi_config.DISPLAY_WIDTH,
            height=europi_config.DISPLAY_HEIGHT,
            sda=europi_config.DISPLAY_SDA,
            scl=europi_config.DISPLAY_SCL,
            channel=europi_config.DISPLAY_CHANNEL,
            freq=europi_config.DISPLAY_FREQUENCY,
            contrast=europi_config.DISPLAY_CONTRAST,
            rotate=europi_config.ROTATE_DISPLAY,
        )
    except Exception as err:
        log_warning(
            f"Failed to initialize display: {err}. Is the hardware connected properly?", "europi"
        )
        oled = DummyDisplay(
            width=europi_config.DISPLAY_WIDTH,
            height=europi_config.DISPLAY_HEIGHT,
        )
else:
    log_warning("No display hardware detected; falling back to DummyDisplay", "europi")
    oled = DummyDisplay(
        width=europi_config.DISPLAY_WIDTH,
        height=europi_config.DISPLAY_HEIGHT,
    )

# Connect to wifi, if supported
if europi_config.PICO_MODEL == MODEL_PICO_W or europi_config.PICO_MODEL == MODEL_PICO_2W:
    try:
        oled.centre_text(
            f"""WiFi connecting
{experimental_config.WIFI_SSID}
Cancel: B1"""
        )
        wifi_connection = WifiConnection()
    except WifiError as err:
        wifi_connection = None
else:
    wifi_connection = None

# Reset the module state upon import.
reset_state()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface for interacting with realtime clock hardware.

For consistency, we always assume that the external clock source is synchronized with UTC. This means
your I2C clock should be set to UTC time, not local time. If you set your I2C clock to local time
some scripts may behave incorrectly.

The Raspberry Pi Pico (and official variants like the Pico W, Pico 2, etc...) does NOT
include a realtime clock. All RTC implementations rely on some external reference time, e.g.
- external hardware (e.g. an I2C-supported external clock module)
- a wireless connection and an accessible NTP server

The external clock source must implement the ExternalClockSource class
"""


class ExternalClockSource:
    """
    A generic class representing any external, canonical clock source.

    Any network- or I2C-based external time source should inherit from this class and
    implement the relevant functions.

    The implemented clock source must provide the time and date in UTC, 24-hour time.
    """

    YEAR = 0
    MONTH = 1
    DAY = 2
    HOUR = 3
    MINUTE = 4
    SECOND = 5
    WEEKDAY = 6
    YEARDAY = 7

    # fmt: off
    # The lengths of the months in a non-leap-year
    _month_lengths = [
        31,
        28,
        31,
        30,
        31,
        30,
        31,
        31,
        30,
        31,
        30,
        31
    ]
    # fmt: on

    def __init__(self):
        pass

    def datetime(self):
        """
        Get the current UTC time as a tuple.

        see: https://docs.micropython.org/en/latest/library/time.html#time.localtime

        :return: a tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        raise NotImplementedError()

    def set_datetime(self, datetime):
        """
        Set the clock's current UTC time.

        If the clock does not support setting (e.g. it's an NTP source we can only read from)
        your sub-class should implement this method anyway and simply pass.

        see: https://docs.micropython.org/en/latest/library/time.html#time.localtime

        :param datetime:  A tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        raise NotImplementedError()

    def is_leap_year(self, datetime):
        """
        Determine if the datetime's year is a leap year or not.

        :return:  True if the datetime is a leap year, otherwise False
        """
        # a year is a leap year if it is divisible by 4
        # but NOT a multple of 100, unless it's also a multiple of 400
        year = datetime[self.YEAR]
        return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

    def year_length(self, datetime):
        """
        Determine the number of days in the datetime's year.

        :return:  The number of days in the year, taking leap years into account
        """
        if self.is_leap_year(datetime):
            return 366
        return 365

    def month_length(self, datetime):
        """
        Get the numer of days in the month.

        This takes leap-years into consideration

        :return:  The number of days in the datetime's month
        """
        if datetime[self.MONTH] == 2 and self.is_leap_year(datetime):
            return 29
        return self._month_lengths[datetime[self.MONTH] - 1]

    def check_valid_datetime(self, datetime):
        """
        Check if a datetime contains valid values.

        :raises ValueError: if any field is out of range

        :param datetime:  The datetime tuple to validate
        """
        # fmt: off
        n_days = self.month_length(datetime)
        if (
            # To anyone reading this from the past: congrats on time-travel! Your year is not supported
            # To anyone reading this from the future: sorry, you'll need to modify some code to get your
            # dates supported
            # Also, it's pretty cool if you're actually reading this beyond 2099. I'm flattered.
            datetime[self.YEAR] < 2000 or
            datetime[self.YEAR] > 2099 or  # somewhat arbitrary, but the DS3231 library I'm using cuts off here, too
            datetime[self.MONTH] < 1 or
            datetime[self.MONTH] > 12 or
            datetime[self.DAY] < 1 or
            datetime[self.DAY] > n_days or
            datetime[self.HOUR] < 0 or
            datetime[self.HOUR] >= 24 or  # time is 00:00:00 to 23:59:59
            datetime[self.MINUTE] < 0 or
            datetime[self.MINUTE] >= 60 or
            (
                # seconds are optional, must be 0-59
                len(datetime) >= self.SECOND + 1 and
                (
                    datetime[self.SECOND] < 0 or
                    datetime[self.SECOND] >= 60
                )
            ) or
            (
                # weekday is optional, must be 1-7
                len(datetime) >= self.WEEKDAY + 1 and
                (
                    datetime[self.WEEKDAY] < 1 or
                    datetime[self.WEEKDAY] > 7
                )
            )
        ):
            raise ValueError("Invalid datetime tuple")
        # fmt: on



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides a base class for scripts which wish to participate in the bootloader menu."""

import os
import json
from utime import ticks_diff, ticks_ms
from configuration import ConfigSpec, ConfigFile
from europi_config import EuroPiConfig
from file_utils import load_file, delete_file, load_json_file


class EuroPiScript:
    """A base class for scripts which wish to participate in the bootloader menu and add save state functionality.

    **Menu Inclusion**

    To make your script compatible with the menu, you must:

    #. import ``EuroPiScript``
    #. define your script as a class that inherits from this base class.
    #. override ``main()`` to define your script's main loop
    #. Surround the call to ``main()`` with a guard to prevent execution upon import

    An minimal example::

       from europi import oled
       from europi_script import EuroPiScript  # 1

       class HelloWorld(EuroPiScript):  # 2
           def main():  # 3
               oled.centre_text("Hello world")


       if __name__ == "__main__":  # 4
           HelloWorld().main()

    To include your script in the menu it must be added to the ``EUROPI_SCRIPTS`` list in ``contrib/menu.py``.

    **Save/Load Script State**

    Optionally, you can add a bit of code to enable your script to save state, and load previous state at startup. By
    default, when exiting the script to menu selection, ``save_state()`` will be called. Additionally, you can add calls
    to ``save_state()`` whenever state changes.

    When adding ``save_state()`` calls to your script, there are a few important considerations to keep in mind:

        * Frequency of saves - scripts should only save state to disk when state changes, and should not save too frequently
          because os write operations are expensive in terms of time. Saving too frequently will affect the performance of a script.

        * Save state file size - The pico only has about 1MB of free space available so save state storage format is important to
          keep as minimal as possible.

        * No externally influenced input - The instance variables your script saves should not be externally influenced,
          meaning you should not save the current knob position, current analog input value or current digital input value.

    To add the ability to save and load state, you must:

        #. **Initialize base classes** When implementing the ``EuroPiScript`` base class, its initialization method must be
           called to initialize its intance variables.

        #. **Call the inherited EuroPiScript method load_state_X().** The ``EuroPiScript`` base class has ``load_state_X()``
           methods to check for a previously saved state of a specific format. When initializing your script, call ``load_state_X()``
           where ``X`` is the persistance format of choice. If no state is found, an empty value will be returned.

        #. **Apply saved state variables to this instance.** Set state variables with default fallback values if not found in the json save state.

        #. **Save state upon state change.** When a state variable changes, call the save state function.

        #. **Implement save_state() method.** Provide an implementation to serialize the state variables into a string, JSON, or
           bytes an call the appropriate save state method.

        #. **Throttle the frequency of saves.** Saving state too often could negatively impact the performance of your script, so it is
           advised to add some checks in your code to ensure it doesn't save too frequently.


    Here is an extension of the script above with some added trivial features that incorporate saving and loading script state::

        from europi import oled
        from europi_script import EuroPiScript

        class HelloWorld(EuroPiScript):
            def __init__(self):
                super().__init__()  # 1

                state = self.load_state_json()  # 2

                self.counter = state.get("counter", 0)  # 3
                self.enabled = state.get("enabled", True)

                @din.handler
                def increment_counter():
                    if self.enabled:
                        self.counter += 1
                        self.save_state()  # 4

                @b1.handler
                def toggle_enablement():
                    self.enabled = not self.enabled
                    self.save_state()  # 4

            def save_state(self):  # 5
                # Don't save if it has been less than 5 seconds since last save.
                if self.last_saved() < 5000:  # 6
                    return

                state = {
                    "counter": self.counter,
                    "enabled": self.enabled,
                }
                self.save_state_json(state)

            def main(self):
                oled.centre_text("Hello world")


    .. note::
       EuroPiScripts should not call ``europi.reset_state()`` as this call would remove the button handlers that
       allow the user to exit the program and return to the menu. Similarly, EuroPiScripts should not override the
       handlers that provide this functionality.

    **Configuration**

    Optionally, you can add code to your script to allow for configuration of behavior outside of
    the hardware UI. A script's configuration options are specified by the `config_points()` method.
    The method returns a list of `ConfigPoint` objects, each of which describes a config point,
    including its name, type, and valid values. See the method's docstrings for details. For
    example, our hello world script could configure its language::

        @classmethod
        def config_points(cls):
            return [configuration.choice(name="language", choices=["english", "french"], default="english")]

    Our main method could then use the value of this configuration to display its greeting in the
    configured language:

        def main(self):
            if self.config.LANGUAGE == "french":
                oled.centre_text("Bonjour le monde")
            else:
                oled.centre_text("Hello world")

    Configuration files are validated, so scripts do not need to worry about invalid values. Validation
    failures raise exceptions with messages that will help the user correct their configurations.

    Users can create and edit configuration files in order to change a script's configuration. The
    files should be uploaded to the pico in the `/config` directory. To assist in generating initial
    versions of these files, see `/scripts/generate_default_configs.py`.
    """

    def __init__(self):
        self._last_saved = 0
        self.config = EuroPiScript._load_config_for_class(self.__class__)
        self.europi_config = EuroPiScript._load_config_for_class(EuroPiConfig)

    def main(self):
        """Override this method with your script's main loop method."""
        raise NotImplementedError

    @classmethod
    def display_name(cls) -> str:
        """Returns the string used to identify this script in the Menu. Defaults to the class name. Override it if you
        would like to use a different name::

           @classmethod
           def display_name(cls):
               return "Hello World"

        Note that the screen is only 16 characters wide. Anything longer will be cut off.
        """
        return cls.__qualname__

    # Save State Methods

    @property
    def _state_filename(self):
        return f"saved_state_{self.__class__.__qualname__}.txt"

    def save_state(self):
        """Encode state and call the appropriate persistence save method.

        Override this class with the script specific encoding of state into
        the persistence format that best matches its use case. Then call the
        appropriate save method, such as `save_state_json(state)`. See the
        class documentation for a full example.
        """
        self.save_state_json({})

    def save_state_bytes(self, state: bytes):
        """Take state in persistence format as bytes and write to disk.

        .. note::
            Be mindful of how often `save_state_bytes()` is called because
            writing to disk too often can slow down the performance of your
            script. Only call save state when state has changed and consider
            adding a time since last save check to reduce save frequency.
        """
        with open(self._state_filename, "wb") as file:
            file.write(state)
            self._last_saved = ticks_ms()

    def save_state_json(self, state: dict):
        """Take state as a dict and save as a json string.

        .. note::
            Be mindful of how often `save_state_json()` is called because
            writing to disk too often can slow down the performance of your
            script. Only call save state when state has changed and consider
            adding a time since last save check to reduce save frequency.
        """
        with open(self._state_filename, "w") as file:
            json.dump(state, file, separators=(",\n", ":"))
            self._last_saved = ticks_ms()

    def load_state_bytes(self) -> bytes:
        """Check disk for saved state, if it exists, return the raw state value as bytes.

        Check for a previously saved state. If it exists, return state as a
        byte string. If no state is found, an empty string will be returned.
        """
        return load_file(self._state_filename, "rb")

    def load_state_json(self) -> dict:
        """Load previously saved state as a dict.

        Check for a previously saved state. If it exists, return state as a
        dict. If no state is found, an empty dictionary will be returned.
        """
        return load_json_file(self._state_filename)

    def remove_state(self):
        """Remove the state file for this script."""
        delete_file(self._state_filename)

    def last_saved(self):
        """Return the ticks in milliseconds since last save."""
        try:
            return ticks_diff(ticks_ms(), self._last_saved)
        except AttributeError:
            raise Exception("EuroPiScript classes must call `super().__init__()`.")

    # config methods

    @classmethod
    def config_points(cls) -> "List[ConfigPoint]":
        """Returns a list of ConfigPoints describing this script's configuration options. By default
        this function returns an empty list. Override it if you'd like configuration points."""
        return []

    @staticmethod
    def _load_config_for_class(cls):
        return ConfigFile.load_config(cls, ConfigSpec(cls.config_points()))
